# Copyright 2023 Cognite AS
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import re
import sys
import time
from collections import defaultdict
from collections.abc import Hashable, Iterable, Sequence
from graphlib import CycleError, TopologicalSorter
from pathlib import Path
from time import sleep
from typing import Any, final

from cognite.client.data_classes import filters
from cognite.client.data_classes.capabilities import (
    Capability,
    DataModelInstancesAcl,
    DataModelsAcl,
)
from cognite.client.data_classes.data_modeling import ContainerId, DataModelId, ViewId
from rich import print
from rich.console import Console
from rich.markup import escape
from rich.panel import Panel

from cognite_toolkit._cdf_tk import constants
from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.client.http_client import ToolkitAPIError
from cognite_toolkit._cdf_tk.client.request_classes.filters import (
    ContainerFilter,
    DataModelFilter,
    InstanceFilter,
    ViewFilter,
)
from cognite_toolkit._cdf_tk.client.resource_classes.data_modeling import (
    ContainerReference,
    ContainerRequest,
    ContainerResponse,
    DataModelReference,
    DataModelRequest,
    DataModelResponse,
    DirectNodeRelation,
    EdgeRequest,
    EdgeResponse,
    NodeRequest,
    NodeResponse,
    RequiresConstraintDefinition,
    SpaceReference,
    SpaceRequest,
    SpaceResponse,
    ViewCorePropertyResponse,
    ViewReference,
    ViewRequest,
    ViewResponse,
)
from cognite_toolkit._cdf_tk.client.resource_classes.data_modeling._instance import InstanceSlimDefinition
from cognite_toolkit._cdf_tk.client.resource_classes.graphql_data_model import (
    GraphQLDataModelRequest,
    GraphQLDataModelResponse,
)
from cognite_toolkit._cdf_tk.client.resource_classes.instance_api import (
    TypedEdgeIdentifier,
    TypedNodeIdentifier,
    TypedViewReference,
)
from cognite_toolkit._cdf_tk.constants import BUILD_FOLDER_ENCODING, HAS_DATA_FILTER_LIMIT
from cognite_toolkit._cdf_tk.cruds._base_cruds import (
    ResourceContainerCRUD,
    ResourceCRUD,
)
from cognite_toolkit._cdf_tk.exceptions import GraphQLParseError, ToolkitCycleError, ToolkitFileNotFoundError
from cognite_toolkit._cdf_tk.resource_classes import (
    ContainerYAML,
    DataModelYAML,
    EdgeYAML,
    GraphQLDataModelYAML,
    NodeYAML,
    SpaceYAML,
    ViewYAML,
)
from cognite_toolkit._cdf_tk.tk_warnings import HighSeverityWarning, LowSeverityWarning, MediumSeverityWarning
from cognite_toolkit._cdf_tk.utils import (
    GraphQLParser,
    calculate_hash,
    in_dict,
    load_yaml_inject_variables,
    quote_int_value_by_key_in_yaml,
    safe_read,
    sanitize_filename,
    to_diff,
)
from cognite_toolkit._cdf_tk.utils.diff_list import diff_list_identifiable, dm_identifier

from .auth import GroupAllScopedCRUD


@final
class SpaceCRUD(ResourceContainerCRUD[SpaceReference, SpaceRequest, SpaceResponse]):
    item_name = "nodes and edges"
    folder_name = "data_modeling"
    resource_cls = SpaceResponse
    resource_write_cls = SpaceRequest
    kind = "Space"
    yaml_cls = SpaceYAML
    dependencies = frozenset({GroupAllScopedCRUD})
    _doc_url = "Spaces/operation/ApplySpaces"
    delete_recreate_limit_seconds: int = 10

    def __init__(self, client: ToolkitClient, build_dir: Path | None, console: Console | None) -> None:
        super().__init__(client, build_dir, console)
        self._deleted_time_by_id: dict[SpaceReference, float] = {}

    @property
    def display_name(self) -> str:
        return "spaces"

    @classmethod
    def get_required_capability(
        cls, items: Sequence[SpaceRequest] | None, read_only: bool
    ) -> list[Capability] | list[Capability]:
        if not items and items is not None:
            return []

        actions = [DataModelsAcl.Action.Read] if read_only else [DataModelsAcl.Action.Read, DataModelsAcl.Action.Write]

        return [DataModelsAcl(actions, DataModelsAcl.Scope.All())]

    @classmethod
    def get_id(cls, item: SpaceRequest | SpaceResponse | dict) -> SpaceReference:
        if isinstance(item, dict):
            return SpaceReference(space=item["space"])
        return item.as_id()

    @classmethod
    def dump_id(cls, id: SpaceReference) -> dict[str, Any]:
        return id.dump()

    @classmethod
    def as_str(cls, id: SpaceReference) -> str:
        return sanitize_filename(id.space)

    def dump_resource(self, resource: SpaceResponse, local: dict[str, Any] | None = None) -> dict[str, Any]:
        dumped = resource.as_request_resource().dump()
        has_local = local is not None
        local = local or {}
        for key in ["description", "name"]:
            if has_local and dumped.get(key) is None and key not in local:
                # Set to null by server.
                dumped.pop(key, None)
        return dumped

    def create(self, items: Sequence[SpaceRequest]) -> list[SpaceResponse]:
        for item in items:
            item_id = self.get_id(item)
            if item_id in self._deleted_time_by_id:
                elapsed_since_delete = time.perf_counter() - self._deleted_time_by_id[item_id]
                if elapsed_since_delete < self.delete_recreate_limit_seconds:
                    time.sleep(self.delete_recreate_limit_seconds - elapsed_since_delete)
        return self.client.tool.spaces.create(items)

    def retrieve(self, ids: Sequence[SpaceReference]) -> list[SpaceResponse]:
        return self.client.tool.spaces.retrieve(list(ids))

    def update(self, items: Sequence[SpaceRequest]) -> list[SpaceResponse]:
        return self.create(items)

    def delete(self, ids: Sequence[SpaceReference]) -> int:
        existing = self.client.tool.spaces.retrieve(list(ids))
        is_global = {space.space for space in existing if space.is_global}
        if is_global:
            print(
                f"  [bold yellow]WARNING:[/] Spaces {list(is_global)} are global and cannot be deleted, skipping delete, for these."
            )
        to_delete = [space_ref for space_ref in ids if space_ref.space not in is_global]
        self.client.tool.spaces.delete(to_delete)
        for item_id in to_delete:
            self._deleted_time_by_id[item_id] = time.perf_counter()
        return len(to_delete)

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: Sequence[Hashable] | None = None,
    ) -> Iterable[SpaceResponse]:
        if space:
            yield from self.client.tool.spaces.retrieve([SpaceReference(space=space)])
        else:
            for batch in self.client.tool.spaces.iterate():
                yield from batch

    def count(self, ids: Sequence[SpaceReference]) -> int:
        # Bug in spec of aggregate requiring view_id to be passed in, so we cannot use it.
        # When this bug is fixed, it will be much faster to use aggregate.
        spaces = [space_ref.space for space_ref in ids]

        return sum(len(batch) for batch in self._iterate_over_nodes(spaces)) + sum(
            len(batch) for batch in self._iterate_over_edges(spaces)
        )

    def drop_data(self, ids: Sequence[SpaceReference]) -> int:
        spaces = [space_ref.space for space_ref in ids]
        if not spaces:
            return 0
        print(f"[bold]Deleting existing data in spaces {ids}...[/]")
        nr_of_deleted = 0
        for edge_ids in self._iterate_over_edges(spaces):
            self.client.tool.instances.delete(edge_ids)
            nr_of_deleted += len(edge_ids)
        for node_ids in self._iterate_over_nodes(spaces):
            self.client.tool.instances.delete(node_ids)
            nr_of_deleted += len(node_ids)
        return nr_of_deleted

    def _iterate_over_nodes(self, spaces: list[str]) -> Iterable[list[TypedNodeIdentifier]]:
        if not spaces:
            return
        filter_ = InstanceFilter(instance_type="node", space=spaces)
        for instances in self.client.tool.instances.iterate(filter=filter_):
            yield [inst.as_id() for inst in instances]  # type: ignore[misc]

    def _iterate_over_edges(self, spaces: list[str]) -> Iterable[list[TypedEdgeIdentifier]]:
        if not spaces:
            return
        filter_ = InstanceFilter(instance_type="edge", space=spaces)
        for instances in self.client.tool.instances.iterate(filter=filter_):
            yield [inst.as_id() for inst in instances]  # type: ignore[misc]


class ContainerCRUD(ResourceContainerCRUD[ContainerReference, ContainerRequest, ContainerResponse]):
    item_name = "nodes and edges"
    folder_name = "data_modeling"
    resource_cls = ContainerResponse
    resource_write_cls = ContainerRequest
    kind = "Container"
    dependencies = frozenset({SpaceCRUD})
    yaml_cls = ContainerYAML
    _doc_url = "Containers/operation/ApplyContainers"
    sub_folder_name = "containers"

    def __init__(
        self,
        client: ToolkitClient,
        build_dir: Path | None,
        console: Console | None = None,
        topological_sort_implements: bool = False,
    ) -> None:
        super().__init__(client, build_dir, console)
        self._container_by_id: dict[ContainerReference, ContainerResponse] = {}

    @property
    def display_name(self) -> str:
        return "containers"

    @classmethod
    def get_required_capability(
        cls, items: Sequence[ContainerRequest] | None, read_only: bool
    ) -> Capability | list[Capability]:
        if not items and items is not None:
            return []

        actions = [DataModelsAcl.Action.Read] if read_only else [DataModelsAcl.Action.Read, DataModelsAcl.Action.Write]

        scope = (
            DataModelsAcl.Scope.SpaceID(list({item.space for item in items}))
            if items is not None
            else DataModelsAcl.Scope.All()
        )

        return DataModelsAcl(actions, scope)

    @classmethod
    def get_id(cls, item: ContainerRequest | ContainerResponse | dict) -> ContainerReference:
        if isinstance(item, dict):
            if missing := tuple(k for k in {"space", "externalId"} if k not in item):
                # We need to raise a KeyError with all missing keys to get the correct error message.
                raise KeyError(*missing)
            return ContainerReference(space=item["space"], external_id=item["externalId"])
        return item.as_id()

    @classmethod
    def dump_id(cls, id: ContainerReference) -> dict[str, Any]:
        return id.dump()

    @classmethod
    def get_dependent_items(cls, item: dict) -> Iterable[tuple[type[ResourceCRUD], Hashable]]:
        if "space" in item:
            yield SpaceCRUD, SpaceReference(space=item["space"])
        # Note that we are very careful in the code below to not raise an exception if the
        # item is not properly formed. If that is the case, an appropriate warning will be given elsewhere.
        for prop in item.get("properties", {}).values():
            if not isinstance(prop, dict):
                continue
            prop_type = prop.get("type", {})
            if isinstance(prop_type, dict) and prop_type.get("type") == "direct":
                if isinstance(prop_type.get("container"), dict):
                    container = prop_type["container"]
                    if "space" in container and "externalId" in container and container.get("type") == "container":
                        yield (
                            ContainerCRUD,
                            ContainerReference(space=container["space"], external_id=container["externalId"]),
                        )

    def dump_resource(self, resource: ContainerResponse, local: dict[str, Any] | None = None) -> dict[str, Any]:
        dumped = resource.as_request_resource().dump()
        has_local = local is not None
        local = local or {}
        for key in ["description", "name"]:
            if has_local and dumped.get(key) is None and key not in local:
                # Set to null by server.
                dumped.pop(key, None)

        for key in ["constraints", "indexes"]:
            if not dumped.get(key) and key not in local:
                # Set to empty dict by server.
                dumped.pop(key, None)
                continue
            if isinstance((cdf_value := dumped.get(key)), dict) and isinstance((local_value := local.get(key)), dict):
                for cdf_id, cdf_item in cdf_value.items():
                    local_item = local_value.get(cdf_id)
                    if (
                        isinstance(local_item, dict)
                        and "bySpace" not in local_item
                        and isinstance(cdf_item, dict)
                        and cdf_item.get("bySpace") is False
                    ):
                        cdf_item.pop("bySpace", None)
        local_prop_by_id = local.get("properties", {})
        for prop_id, cdf_prop in dumped.get("properties", {}).items():
            if prop_id not in local_prop_by_id:
                continue
            local_prop = local_prop_by_id[prop_id]
            for key, default in [("immutable", False), ("autoIncrement", False), ("nullable", True)]:
                if has_local and cdf_prop.get(key) is default and key not in local_prop:
                    cdf_prop.pop(key, None)
            cdf_type = cdf_prop.get("type", {})
            local_type = local_prop.get("type", {})
            for key, type_default in [("list", False), ("collation", "ucs_basic")]:
                if has_local and cdf_type.get(key) == type_default and key not in local_type:
                    cdf_type.pop(key, None)
            if has_local and "usedFor" not in local and dumped.get("usedFor") == "node":
                # Only drop if set to default by server.
                dumped.pop("usedFor", None)
        return dumped

    def create(self, items: Sequence[ContainerRequest]) -> list[ContainerResponse]:
        return self.client.tool.containers.create(items)

    def retrieve(self, ids: Sequence[ContainerReference]) -> list[ContainerResponse]:
        return self.client.tool.containers.retrieve(list(ids))

    def update(self, items: Sequence[ContainerRequest]) -> list[ContainerResponse]:
        updated = self.create(items)
        # The API might silently fail to update a container.
        updated_by_id = {item.as_id(): item for item in updated}
        for local in items:
            item_id = local.as_id()
            local_dict = local.dump()
            if item_id not in updated_by_id:
                raise ToolkitAPIError(
                    f"The container {item_id} was not updated. You might need to delete and recreate it.",
                    code=500,
                )
            cdf_dict = self.dump_resource(updated_by_id[item_id], local_dict)
            if cdf_dict != local_dict:
                is_verbose = "-v" in sys.argv or "--verbose" in sys.argv
                if is_verbose:
                    print(
                        Panel(
                            "\n".join(to_diff(cdf_dict, local_dict)),
                            title=f"{self.display_name}: {item_id}",
                            expand=False,
                        )
                    )
                suffix = "" if is_verbose else " (use -v for more info)"
                HighSeverityWarning(
                    f"The container {item_id} was not updated. You might need to delete and recreate it{suffix}."
                ).print_warning()
        return updated

    def delete(self, ids: Sequence[ContainerReference]) -> int:
        self.client.tool.containers.delete(list(ids))
        return len(ids)

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: Sequence[Hashable] | None = None,
    ) -> Iterable[ContainerResponse]:
        for batch in self.client.tool.containers.iterate(filter=ContainerFilter(space=space) if space else None):
            yield from batch

    def count(self, ids: Sequence[ContainerReference]) -> int:
        # Bug in spec of aggregate requiring view_id to be passed in, so we cannot use it.
        # When this bug is fixed, it will be much faster to use aggregate.
        existing_containers = self.client.tool.containers.retrieve(list(ids))
        return sum(len(batch) for batch in self._iterate_over_nodes(existing_containers)) + sum(
            len(batch) for batch in self._iterate_over_edges(existing_containers)
        )

    def drop_data(self, ids: Sequence[ContainerReference]) -> int:
        nr_of_deleted = 0
        existing_containers = self.client.tool.containers.retrieve(list(ids))
        for node_ids in self._iterate_over_nodes(existing_containers):
            self.client.tool.instances.delete(node_ids)
            nr_of_deleted += len(node_ids)
        for edge_ids in self._iterate_over_edges(existing_containers):
            self.client.tool.instances.delete(edge_ids)
            nr_of_deleted += len(edge_ids)
        return nr_of_deleted

    def _iterate_over_nodes(self, containers: list[ContainerResponse]) -> Iterable[list[TypedNodeIdentifier]]:
        container_ids = [container.as_id() for container in containers if container.used_for in ["node", "all"]]
        if not container_ids:
            return
        for container_id_chunk in self._chunker(container_ids, HAS_DATA_FILTER_LIMIT):
            is_container = filters.HasData(
                containers=[ContainerId(space=cid.space, external_id=cid.external_id) for cid in container_id_chunk]
            )
            for instances in self.client.data_modeling.instances(
                chunk_size=1000, instance_type="node", filter=is_container, limit=-1
            ):
                yield [TypedNodeIdentifier(space=nid.space, external_id=nid.external_id) for nid in instances.as_ids()]

    def _iterate_over_edges(self, containers: list[ContainerResponse]) -> Iterable[list[TypedEdgeIdentifier]]:
        container_ids = [container.as_id() for container in containers if container.used_for in ["edge", "all"]]
        if not container_ids:
            return

        for container_id_chunk in self._chunker(container_ids, HAS_DATA_FILTER_LIMIT):
            is_container = filters.HasData(
                containers=[ContainerId(space=cid.space, external_id=cid.external_id) for cid in container_id_chunk]
            )
            for instances in self.client.data_modeling.instances(
                chunk_size=1000, instance_type="edge", limit=-1, filter=is_container
            ):
                yield [TypedEdgeIdentifier(space=eid.space, external_id=eid.external_id) for eid in instances.as_ids()]

    def _lookup_containers(
        self, container_ids: Sequence[ContainerReference]
    ) -> dict[ContainerReference, ContainerResponse]:
        ids_to_lookup = [container_id for container_id in container_ids if container_id not in self._container_by_id]
        if ids_to_lookup:
            retrieved_containers = self.client.tool.containers.retrieve(ids_to_lookup)
            for container in retrieved_containers:
                self._container_by_id[container.as_id()] = container
        if missing_container_ids := set(container_ids) - set(self._container_by_id.keys()):
            MediumSeverityWarning(
                f"Containers {missing_container_ids} not found or you don't have permission to access them."
            ).print_warning(console=self.console)
        return {
            container_id: self._container_by_id[container_id]
            for container_id in container_ids
            if container_id in self._container_by_id
        }

    def _find_direct_container_dependencies(
        self, container_ids: Sequence[ContainerReference]
    ) -> dict[ContainerReference, set[ContainerReference]]:
        containers_by_id = self._lookup_containers(container_ids)
        container_dependencies: dict[ContainerReference, set[ContainerReference]] = defaultdict(set)
        for container_id, container in containers_by_id.items():
            for constraint in (container.constraints or {}).values():
                if not isinstance(constraint, RequiresConstraintDefinition):
                    continue
                container_dependencies[container_id].add(constraint.require)
            for property in container.properties.values():
                if not isinstance(property.type, DirectNodeRelation) or property.type.container is None:
                    continue
                container_dependencies[container_id].add(property.type.container)
        return container_dependencies

    def _propagate_indirect_container_dependencies(
        self,
        container_dependencies_by_id: dict[ContainerReference, set[ContainerReference]],
        dependants: Sequence[ContainerReference],
    ) -> dict[ContainerReference, set[ContainerReference]]:
        """Propagate indirect container dependencies using a recursive approach.

        Args:
            container_dependencies_by_id: Mapping of container IDs to their direct dependencies
            dependants: Chain of dependant containers to propagate dependencies to

        Returns:
            Updated dictionary mapping each container ID to all its direct and indirect dependencies
        """
        current_container_id = dependants[0]
        dependencies_to_propagate: set[ContainerReference] = set()
        for container_dependency in container_dependencies_by_id[current_container_id]:
            if container_dependency in container_dependencies_by_id:
                # If already processed, propagate its dependencies to current container instead of revisiting it
                dependencies_to_propagate.update(container_dependencies_by_id[container_dependency])
                continue
            self._propagate_indirect_container_dependencies(
                container_dependencies_by_id, [container_dependency, *dependants]
            )
        container_dependencies_by_id[current_container_id].update(dependencies_to_propagate)
        return container_dependencies_by_id

    def _find_direct_and_indirect_container_dependencies(
        self, container_ids: Sequence[ContainerReference]
    ) -> dict[ContainerReference, set[ContainerReference]]:
        container_dependencies_by_id = self._find_direct_container_dependencies(container_ids)
        for container_id in list(container_dependencies_by_id.keys()):
            self._propagate_indirect_container_dependencies(container_dependencies_by_id, [container_id])
        return container_dependencies_by_id

    @staticmethod
    def _chunker(seq: Sequence, size: int) -> Iterable[Sequence]:
        return (seq[pos : pos + size] for pos in range(0, len(seq), size))

    @classmethod
    def as_str(cls, id: ContainerReference) -> str:
        return sanitize_filename(f"{id.space}_{id.external_id}")


class ViewCRUD(ResourceCRUD[ViewReference, ViewRequest, ViewResponse]):
    folder_name = "data_modeling"
    resource_cls = ViewResponse
    resource_write_cls = ViewRequest
    kind = "View"
    dependencies = frozenset({SpaceCRUD, ContainerCRUD})
    yaml_cls = ViewYAML
    _doc_url = "Views/operation/ApplyViews"
    sub_folder_name = "views"

    def __init__(
        self,
        client: ToolkitClient,
        build_dir: Path | None,
        console: Console | None,
        topological_sort_implements: bool = False,
    ) -> None:
        super().__init__(client, build_dir, console)
        self._topological_sort_implements = topological_sort_implements
        self._view_by_id: dict[ViewReference, ViewResponse] = {}

    @property
    def display_name(self) -> str:
        return "views"

    @classmethod
    def get_required_capability(
        cls, items: Sequence[ViewRequest] | None, read_only: bool
    ) -> Capability | list[Capability]:
        if not items and items is not None:
            return []

        actions = [DataModelsAcl.Action.Read] if read_only else [DataModelsAcl.Action.Read, DataModelsAcl.Action.Write]

        scope = (
            DataModelsAcl.Scope.SpaceID(list({item.space for item in items}))
            if items is not None
            else DataModelsAcl.Scope.All()
        )

        return DataModelsAcl(actions, scope)

    @classmethod
    def get_id(cls, item: ViewRequest | ViewResponse | dict) -> ViewReference:
        if isinstance(item, dict):
            if missing := tuple(k for k in {"space", "externalId", "version"} if k not in item):
                # We need to raise a KeyError with all missing keys to get the correct error message.
                raise KeyError(*missing)
            return ViewReference(space=item["space"], external_id=item["externalId"], version=str(item["version"]))

        return item.as_id()

    @classmethod
    def dump_id(cls, id: ViewReference) -> dict[str, Any]:
        return id.dump()

    @classmethod
    def get_dependent_items(cls, item: dict) -> Iterable[tuple[type[ResourceCRUD], Hashable]]:
        if "space" in item:
            yield SpaceCRUD, SpaceReference(space=item["space"])
        if isinstance(implements := item.get("implements", []), list):
            for parent in implements:
                if not isinstance(parent, dict):
                    continue
                if parent.get("type") == "view" and in_dict(["space", "externalId", "version"], parent):
                    yield (
                        ViewCRUD,
                        ViewReference(
                            space=parent["space"],
                            external_id=parent["externalId"],
                            version=str(v) if (v := parent.get("version")) else "",
                        ),
                    )
        for prop in item.get("properties", {}).values():
            if (container := prop.get("container", {})) and container.get("type") == "container":
                if in_dict(("space", "externalId"), container):
                    yield (
                        ContainerCRUD,
                        ContainerReference(space=container["space"], external_id=container["externalId"]),
                    )
            for key, dct_ in [("source", prop), ("edgeSource", prop), ("source", prop.get("through", {}))]:
                if source := dct_.get(key, {}):
                    if source.get("type") == "view" and in_dict(("space", "externalId", "version"), source):
                        yield (
                            ViewCRUD,
                            ViewReference(
                                space=source["space"],
                                external_id=source["externalId"],
                                version=str(v) if (v := source.get("version")) else "",
                            ),
                        )
                    elif source.get("type") == "container" and in_dict(("space", "externalId"), source):
                        yield ContainerCRUD, ContainerReference(space=source["space"], external_id=source["externalId"])

    def safe_read(self, filepath: Path | str) -> str:
        # The version is a string, but the user often writes it as an int.
        # YAML will then parse it as an int, for example, `3_0_2` will be parsed as `302`.
        # This is technically a user mistake, as you should quote the version in the YAML file.
        # However, we do not want to put this burden on the user (knowing the intricate workings of YAML),
        # so we fix it here.
        return quote_int_value_by_key_in_yaml(safe_read(filepath, encoding=BUILD_FOLDER_ENCODING), key="version")

    def dump_resource(self, resource: ViewResponse, local: dict[str, Any] | None = None) -> dict[str, Any]:
        dumped = resource.as_request_resource().dump()
        local = local or {}
        if not dumped.get("properties") and not local.get("properties"):
            if "properties" in local:
                # In case the properties is an empty dict, we still want to keep it in the dump.
                # such that the dumped evaluates to the same as the local.
                dumped["properties"] = local["properties"]
            else:
                dumped.pop("properties", None)
        if not dumped.get("implements") and not local.get("implements"):
            if "implements" in local:
                # In case the implements is an empty list, we still want to keep it in the dump.
                # such that the dumped evaluates to the same as the local.
                dumped["implements"] = local["implements"]
            else:
                dumped.pop("implements", None)
        if resource.implements and len(resource.implements) > 1 and self._topological_sort_implements:
            # This is a special case that we want to do when we run the cdf dump datamodel command.
            # The issue is as follows:
            # 1. If a data model is deployed through GraphQL, the implements for a child view are sorted
            #   from parent, grandparent, etc.
            # 2. If the grand parent has a direct relation that the parent overwrites to update the source.
            #   The child will get the grandparent's source, not the parent's.
            # We sort the implements in topological order to ensure that the child view get the order grandparent,
            # parent, such that the parent's source is used.
            try:
                dumped["implements"] = [
                    view_id.dump() for view_id in self.topological_sort_implements(resource.implements)
                ]
            except ToolkitCycleError as e:
                warning = MediumSeverityWarning(f"Failed to sort implements for view {resource.as_id()}: {e}")
                warning.print_warning(console=self.console)

        local_properties = local.get("properties", {})
        for prop_id, prop in dumped.get("properties", {}).items():
            if prop_id not in local_properties:
                continue
            local_prop = local_properties[prop_id]
            if all(isinstance(v.get("container"), dict) for v in [prop, local_prop]):
                if prop["container"].get("type") == "container" and "type" not in local_prop["container"]:
                    prop["container"].pop("type", None)
            is_connection_prop = "connectionType" in prop
            is_local_connection_prop = "connectionType" in local_prop
            if (
                is_connection_prop
                and is_local_connection_prop
                and "direction" not in local_prop
                and prop.get("direction") == "outwards"
            ):
                # The API will set the direction to outwards by default, so we remove it from the dump.
                prop.pop("direction", None)
            for key, default in [("description", None), ("name", None), ("source", None), ("edgeSource", None)]:
                if prop.get(key) == default and key not in local_prop:
                    prop.pop(key, None)
        return dumped

    def diff_list(
        self, local: list[Any], cdf: list[Any], json_path: tuple[str | int, ...]
    ) -> tuple[dict[int, int], list[int]]:
        if json_path == ("implements",):
            return diff_list_identifiable(local, cdf, get_identifier=dm_identifier)
        return super().diff_list(local, cdf, json_path)

    def create(self, items: Sequence[ViewRequest]) -> list[ViewResponse]:
        try:
            return self.client.tool.views.create(items)
        except ToolkitAPIError as e1:
            if e1.is_auto_retryable:
                # Fallback to creating one by one if the error is auto-retryable.
                return self._fallback_create_one_by_one(items, e1)
            raise

    def _fallback_create_one_by_one(
        self, items: Sequence[ViewRequest], e1: ToolkitAPIError, warn: bool = True
    ) -> list[ViewResponse]:
        if warn:
            MediumSeverityWarning(
                f"Failed to create {len(items)} views error:\n{escape(str(e1))}\n\n----------------------------\nTrying to create one by one..."
            ).print_warning(include_timestamp=True, console=self.console)
        created_list: list[ViewResponse] = []
        for no, item in enumerate(items):
            try:
                created = self.client.tool.views.create([item])
            except ToolkitAPIError as e2:
                raise e2 from e1
            else:
                created_list.extend(created)
        return created_list

    def retrieve(self, ids: Sequence[ViewReference]) -> list[ViewResponse]:
        return self.client.tool.views.retrieve(list(ids), include_inherited_properties=False)

    def update(self, items: Sequence[ViewRequest]) -> list[ViewResponse]:
        return self.create(items)

    def delete(self, ids: Sequence[ViewReference]) -> int:
        to_delete = list(ids)
        nr_of_deleted = 0
        attempt_count = 5
        for attempt_no in range(attempt_count):
            self.client.tool.views.delete(to_delete)
            nr_of_deleted += len(to_delete)
            existing = [view.as_id() for view in self.client.tool.views.retrieve(to_delete)]
            if not existing:
                return nr_of_deleted
            sleep(2)
            to_delete = existing
        else:
            msg = f"  [bold yellow]WARNING:[/] Could not delete views {to_delete} after {attempt_count} attempts."
            if self.console:
                self.console.print(msg)
            else:
                print(msg)
        return nr_of_deleted

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: Sequence[Hashable] | None = None,
    ) -> Iterable[ViewResponse]:
        for batch in self.client.tool.views.iterate(filter=ViewFilter(space=space) if space else None):
            yield from batch

    @classmethod
    def as_str(cls, id: ViewReference) -> str:
        return sanitize_filename(id.external_id)

    def _lookup_views(self, view_ids: list[ViewReference]) -> dict[ViewReference, ViewResponse]:
        """Looks up views by their IDs and caches them."""
        missing_ids = [view_id for view_id in view_ids if view_id not in self._view_by_id]
        if missing_ids:
            retrieved_views = self.client.tool.views.retrieve(missing_ids, include_inherited_properties=True)
            for view in retrieved_views:
                self._view_by_id[view.as_id()] = view
        return {view_id: self._view_by_id[view_id] for view_id in view_ids if view_id in self._view_by_id}

    def get_readonly_properties(self, view_id: ViewReference) -> set[str]:
        """Retrieve the set of read-only properties for a given view."""

        readonly_properties: set[str] = set()

        # Retrieve the view to check its properties
        view = self._lookup_views([view_id]).get(view_id)
        if view is None:
            return readonly_properties

        # Check each property in the view
        for property_identifier, property in view.properties.items():
            if isinstance(
                property, ViewCorePropertyResponse
            ) and property.container_property_identifier in constants.READONLY_CONTAINER_PROPERTIES.get(
                property.container.as_tuple(), set()
            ):
                readonly_properties.add(property_identifier)
        return readonly_properties

    def _build_view_implements_dependencies(
        self, view_by_ids: dict[ViewReference, ViewResponse], include: set[ViewReference] | None = None
    ) -> dict[ViewReference, set[ViewReference]]:
        """Build a dependency graph based on view implements relationships.

        Args:
            view_by_ids: Mapping of view IDs to ViewResponse objects
            include: Optional set of view IDs to include in the dependencies, if None, all views are included.

        Returns:
            Dictionary mapping each view ID to the set of view IDs it depends on (implements)
        """
        dependencies: dict[ViewReference, set[ViewReference]] = {}
        for view_id, view in view_by_ids.items():
            dependencies[view_id] = set()
            for implemented_view_id in view.implements or []:
                if include is None or implemented_view_id in include:
                    dependencies[view_id].add(implemented_view_id)
        return dependencies

    def topological_sort_implements(self, view_ids: list[ViewReference]) -> list[ViewReference]:
        """Sorts the views in topological order based on their implements and through properties."""
        view_by_ids = self._lookup_views(view_ids)
        parents_by_child = self._build_view_implements_dependencies(view_by_ids)

        try:
            sorted_views = list(TopologicalSorter(parents_by_child).static_order())
        except CycleError as e:
            raise ToolkitCycleError(
                f"Failed to sort views topologically. This likely due to a cycle in implements. {e.args[1]}"
            )

        return sorted_views

    def topological_sort_container_constraints(
        self, view_ids: list[ViewReference]
    ) -> tuple[list[ViewReference], list[ViewReference]]:
        """Sorts the views in topological order based on their container constraints.

        Returns:
            A tuple containing the sorted views and cyclic views that could not be sorted (if any).
        """

        view_by_ids = self._lookup_views(view_ids)
        if missing_view_ids := set(view_ids) - set(view_by_ids.keys()):
            MediumSeverityWarning(
                f"Views {missing_view_ids} not found or you don't have permission to access them, skipping dependency check."
            ).print_warning(console=self.console)
            return view_ids, []

        view_to_containers: dict[ViewReference, set[ContainerReference]] = {}
        container_to_views: defaultdict[ContainerReference, set[ViewReference]] = defaultdict(set)
        for view_id, view in view_by_ids.items():
            view_to_containers[view_id] = set(view.mapped_containers)
            for container_id in view_to_containers[view_id]:
                container_to_views[container_id].add(view_id)

        container_crud = ContainerCRUD.create_loader(self.client)
        container_dependencies_by_id = container_crud._find_direct_and_indirect_container_dependencies(
            list(container_to_views.keys())
        )

        # First, add view dependencies based on implements relationships
        view_dependencies = self._build_view_implements_dependencies(view_by_ids, set(view_to_containers.keys()))

        # Then, add view dependencies based on mapped container constraints
        for view_id, mapped_containers in view_to_containers.items():
            for container_id in mapped_containers:
                # Get all containers this container depends on
                if container_id not in container_dependencies_by_id:
                    continue
                for required_container in container_dependencies_by_id[container_id]:
                    if required_container not in container_to_views:
                        continue
                    # If this view already implements the required container, the requirement is self-satisfied
                    # and we don't need to depend on other views that also implement it (they are peers).
                    if required_container in mapped_containers:
                        continue
                    # This view doesn't implement the required container, so depend on all views that do
                    view_dependencies[view_id].update(container_to_views[required_container])

        cyclic_views: list[ViewReference] = []
        for _ in range(
            len(view_dependencies)
        ):  # Ensure an upper bound on the number of iterations we do when removing cycles.
            try:
                TopologicalSorter(view_dependencies).prepare()
                break
            except CycleError as e:
                cycle_nodes = set(e.args[1])
                cyclic_views.extend(cycle_nodes)
                view_dependencies = {k: v - cycle_nodes for k, v in view_dependencies.items() if k not in cycle_nodes}

        sorted_views = list(TopologicalSorter(view_dependencies).static_order())

        return sorted_views, cyclic_views


@final
class DataModelCRUD(ResourceCRUD[DataModelReference, DataModelRequest, DataModelResponse]):
    folder_name = "data_modeling"
    resource_cls = DataModelResponse
    resource_write_cls = DataModelRequest
    kind = "DataModel"
    dependencies = frozenset({SpaceCRUD, ViewCRUD})
    yaml_cls = DataModelYAML
    _doc_url = "Data-models/operation/createDataModels"

    @property
    def display_name(self) -> str:
        return "data models"

    @classmethod
    def get_required_capability(
        cls, items: Sequence[DataModelRequest] | None, read_only: bool
    ) -> Capability | list[Capability]:
        if not items and items is not None:
            return []

        actions = [DataModelsAcl.Action.Read] if read_only else [DataModelsAcl.Action.Read, DataModelsAcl.Action.Write]

        scope = (
            DataModelsAcl.Scope.SpaceID(list({item.space for item in items}))
            if items is not None
            else DataModelsAcl.Scope.All()
        )

        return DataModelsAcl(actions, scope)

    @classmethod
    def get_id(cls, item: DataModelRequest | DataModelResponse | dict) -> DataModelReference:
        if isinstance(item, dict):
            if missing := tuple(k for k in {"space", "externalId", "version"} if k not in item):
                # We need to raise a KeyError with all missing keys to get the correct error message.
                raise KeyError(*missing)
            return DataModelReference(space=item["space"], external_id=item["externalId"], version=str(item["version"]))
        return item.as_id()

    @classmethod
    def dump_id(cls, id: DataModelReference) -> dict[str, Any]:
        return id.dump()

    @classmethod
    def get_dependent_items(cls, item: dict) -> Iterable[tuple[type[ResourceCRUD], Hashable]]:
        if "space" in item:
            yield SpaceCRUD, SpaceReference(space=item["space"])
        for view in item.get("views", []):
            if in_dict(("space", "externalId"), view):
                yield (
                    ViewCRUD,
                    ViewReference(
                        space=view["space"],
                        external_id=view["externalId"],
                        version=str(v) if (v := view.get("version")) else "",
                    ),
                )

    def safe_read(self, filepath: Path | str) -> str:
        # The version is a string, but the user often writes it as an int.
        # YAML will then parse it as an int, for example, `3_0_2` will be parsed as `302`.
        # This is technically a user mistake, as you should quote the version in the YAML file.
        # However, we do not want to put this burden on the user (knowing the intricate workings of YAML),
        # so we fix it here.
        return quote_int_value_by_key_in_yaml(safe_read(filepath, encoding=BUILD_FOLDER_ENCODING), key="version")

    def dump_resource(self, resource: DataModelResponse, local: dict[str, Any] | None = None) -> dict[str, Any]:
        dumped = resource.as_request_resource().dump()
        local = local or {}
        if "views" not in dumped:
            return dumped
        for key in ["name", "description"]:
            if dumped.get(key) is None and key not in local:
                # Set to null by server.
                dumped.pop(key, None)
        # Sorting in the same order as the local file.
        view_order_by_id = {ViewReference._load(v): no for no, v in enumerate(local.get("views", []))}
        end_of_list = len(view_order_by_id)
        dumped["views"] = sorted(
            dumped["views"], key=lambda v: view_order_by_id.get(ViewReference._load(v), end_of_list)
        )
        return dumped

    def diff_list(
        self, local: list[Any], cdf: list[Any], json_path: tuple[str | int, ...]
    ) -> tuple[dict[int, int], list[int]]:
        if json_path == ("views",):
            return diff_list_identifiable(local, cdf, get_identifier=dm_identifier)
        return super().diff_list(local, cdf, json_path)

    def create(self, items: Sequence[DataModelRequest]) -> list[DataModelResponse]:
        return self.client.tool.data_models.create(items)

    def retrieve(self, ids: Sequence[DataModelReference]) -> list[DataModelResponse]:
        return self.client.tool.data_models.retrieve(list(ids))

    def update(self, items: Sequence[DataModelRequest]) -> list[DataModelResponse]:
        updated = self.create(items)
        # There is a bug in the API not raising an exception if view is removed from a data model.
        # So we check here that the update was fixed.
        updated_by_id = {item.as_id(): item for item in updated}
        for local in items:
            item_id = local.as_id()
            if item_id in updated_by_id:
                views_updated = set(updated_by_id[item_id].views or [])
                views_local = set(local.views or [])
                missing = views_local - views_updated
                extra = views_updated - views_local
                if missing or extra:
                    raise ToolkitAPIError(
                        f"The API did not update the data model, {item_id} correctly. You might have "
                        f"to increase the version number of the data model for it to update.\nMissing views in CDF: {missing}\n"
                        f"Extra views in the CDF: {extra}",
                        code=500,
                    )
            else:
                raise ToolkitAPIError(
                    f"The data model {item_id} was not updated. Please check the data model manually.",
                    code=500,
                )

        return updated

    def delete(self, ids: Sequence[DataModelReference]) -> int:
        self.client.tool.data_models.delete(list(ids))
        return len(ids)

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: Sequence[Hashable] | None = None,
    ) -> Iterable[DataModelResponse]:
        for batch in self.client.tool.data_models.iterate(filter=DataModelFilter(space=space, include_global=False)):
            yield from batch

    @classmethod
    def as_str(cls, id: DataModelReference) -> str:
        return sanitize_filename(id.external_id)


@final
class NodeCRUD(ResourceContainerCRUD[TypedNodeIdentifier, NodeRequest, NodeResponse]):
    item_name = "nodes"
    folder_name = "data_modeling"
    resource_cls = NodeResponse
    resource_write_cls = NodeRequest
    kind = "Node"
    yaml_cls = NodeYAML
    dependencies = frozenset({SpaceCRUD, ViewCRUD, ContainerCRUD})
    _doc_url = "Instances/operation/applyNodeAndEdges"
    sub_folder_name = "nodes"

    def __init__(
        self,
        client: ToolkitClient,
        build_dir: Path | None,
        console: Console | None = None,
        view_id: TypedViewReference | None = None,
    ) -> None:
        super().__init__(client, build_dir, console)
        # View ID is used to retrieve nodes with properties.
        self.view_id = view_id

    @property
    def display_name(self) -> str:
        return "nodes"

    @classmethod
    def get_required_capability(
        cls, items: Sequence[NodeRequest] | None, read_only: bool
    ) -> Capability | list[Capability]:
        if not items and items is not None:
            return []

        actions = (
            [DataModelInstancesAcl.Action.Read]
            if read_only
            else [DataModelInstancesAcl.Action.Read, DataModelInstancesAcl.Action.Write]
        )

        return DataModelInstancesAcl(
            actions,
            DataModelInstancesAcl.Scope.SpaceID(list({item.space for item in items}))
            if items is not None
            else DataModelInstancesAcl.Scope.All(),
        )

    @classmethod
    def get_id(cls, item: NodeRequest | NodeResponse | dict) -> TypedNodeIdentifier:
        if isinstance(item, dict):
            if missing := tuple(k for k in {"space", "externalId"} if k not in item):
                # We need to raise a KeyError with all missing keys to get the correct error message.
                raise KeyError(*missing)
            return TypedNodeIdentifier(space=item["space"], external_id=item["externalId"])
        return item.as_id()

    @classmethod
    def dump_id(cls, id: TypedNodeIdentifier) -> dict[str, Any]:
        return id.dump()

    @classmethod
    def get_dependent_items(cls, item: dict) -> Iterable[tuple[type[ResourceCRUD], Hashable]]:
        if "space" in item:
            yield SpaceCRUD, SpaceReference(space=item["space"])
        for source in item.get("sources", []):
            if (identifier := source.get("source")) and isinstance(identifier, dict):
                if identifier.get("type") == "view" and in_dict(("space", "externalId", "version"), identifier):
                    yield (
                        ViewCRUD,
                        ViewReference(
                            space=identifier["space"],
                            external_id=identifier["externalId"],
                            version=str(v) if (v := identifier.get("version")) else "",
                        ),
                    )
                elif identifier.get("type") == "container" and in_dict(("space", "externalId"), identifier):
                    yield (
                        ContainerCRUD,
                        ContainerReference(space=identifier["space"], external_id=identifier["externalId"]),
                    )

    def dump_resource(self, resource: NodeResponse, local: dict[str, Any] | None = None) -> dict[str, Any]:
        # CDF resource does not have properties set, so we need to do a lookup
        local = local or {}
        sources = [
            TypedViewReference._load(source["source"]) for source in local.get("sources", []) if "source" in source
        ]

        # Default dump
        dumped = resource.as_request_resource().dump()
        if sources:
            try:
                node_id = resource.as_id()
                res = self.client.tool.instances.retrieve([node_id], source=sources[0])
            except ToolkitAPIError:
                ...
            else:
                if res:
                    # Dump with properties populated.
                    dumped = res[0].as_request_resource().dump()

        if "existingVersion" not in local:
            # Existing version is typically not set when creating nodes, but we get it back
            # when we retrieve the node from the server.
            dumped.pop("existingVersion", None)

        if "instanceType" in dumped and "instanceType" not in local:
            # Toolkit uses file suffix to determine instanceType, so we need to remove it from the CDF resource
            # to match the local resource.
            dumped.pop("instanceType")

        return dumped

    def create(self, items: Sequence[NodeRequest]) -> list[InstanceSlimDefinition]:
        return self.client.tool.instances.create(list(items))

    def retrieve(self, ids: Sequence[TypedNodeIdentifier]) -> list[NodeResponse]:
        source_ref = (
            TypedViewReference(
                space=self.view_id.space, external_id=self.view_id.external_id, version=self.view_id.version
            )
            if self.view_id
            else None
        )
        results = self.client.tool.instances.retrieve(list(ids), source=source_ref)
        return [r for r in results if isinstance(r, NodeResponse)]

    def update(self, items: Sequence[NodeRequest]) -> list[InstanceSlimDefinition]:
        return self.client.tool.instances.create(list(items))

    def delete(self, ids: Sequence[TypedNodeIdentifier]) -> int:
        try:
            deleted = self.client.tool.instances.delete(list(ids))
        except ToolkitAPIError as e:
            if "not exist" in str(e) and "space" in str(e).lower():
                return 0
            raise e
        return len(deleted)

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: Sequence[Hashable] | None = None,
    ) -> Iterable[NodeResponse]:
        source_ref = (
            TypedViewReference(
                space=self.view_id.space, external_id=self.view_id.external_id, version=self.view_id.version
            )
            if self.view_id
            else None
        )
        filter_ = InstanceFilter(
            instance_type="node",
            space=[space] if space else None,
            source=source_ref,
        )
        for batch in self.client.tool.instances.iterate(filter=filter_):
            for inst in batch:
                if isinstance(inst, NodeResponse):
                    yield inst

    def count(self, ids: Sequence[TypedNodeIdentifier]) -> int:
        return len(ids)

    def drop_data(self, ids: Sequence[TypedNodeIdentifier]) -> int:
        # Nodes will be deleted in .delete call.
        return 0

    @classmethod
    def as_str(cls, id: TypedNodeIdentifier) -> str:
        return sanitize_filename(f"{id.space}_{id.external_id}")


class GraphQLCRUD(ResourceContainerCRUD[DataModelReference, GraphQLDataModelRequest, GraphQLDataModelResponse]):
    folder_name = "data_modeling"
    resource_cls = GraphQLDataModelResponse
    resource_write_cls = GraphQLDataModelRequest
    kind = "GraphQLSchema"
    dependencies = frozenset({SpaceCRUD, ContainerCRUD})
    item_name = "views"
    yaml_cls = GraphQLDataModelYAML
    _doc_url = "Data-models/operation/createDataModels"
    _hash_name = "CDFToolkitHash:"

    def __init__(self, client: ToolkitClient, build_dir: Path, console: Console | None) -> None:
        super().__init__(client, build_dir, console)
        self._graphql_filepath_cache: dict[DataModelReference, Path] = {}
        self._datamodels_by_view_id: dict[ViewReference, set[DataModelReference]] = defaultdict(set)
        self._dependencies_by_datamodel_id: dict[DataModelReference, set[ViewReference | DataModelReference]] = {}

    @property
    def display_name(self) -> str:
        return "graph QL schemas"

    @classmethod
    def get_id(cls, item: GraphQLDataModelRequest | GraphQLDataModelResponse | dict) -> DataModelReference:
        if isinstance(item, dict):
            if missing := tuple(k for k in {"space", "externalId", "version"} if k not in item):
                # We need to raise a KeyError with all missing keys to get the correct error message.
                raise KeyError(*missing)
            return DataModelReference(space=item["space"], external_id=item["externalId"], version=str(item["version"]))
        return item.as_id()

    @classmethod
    def dump_id(cls, id: DataModelReference) -> dict[str, Any]:
        return id.dump()

    @classmethod
    def get_required_capability(
        cls, items: Sequence[GraphQLDataModelRequest] | None, read_only: bool
    ) -> Capability | list[Capability]:
        if not items and items is not None:
            return []
        actions = [DataModelsAcl.Action.Read] if read_only else [DataModelsAcl.Action.Read, DataModelsAcl.Action.Write]
        return DataModelsAcl(
            actions,
            DataModelsAcl.Scope.SpaceID(list({item.space for item in items}))
            if items is not None
            else DataModelsAcl.Scope.All(),
        )

    @classmethod
    def get_dependent_items(cls, item: dict) -> Iterable[tuple[type[ResourceCRUD], Hashable]]:
        if "space" in item:
            yield SpaceCRUD, SpaceReference(space=item["space"])

    def safe_read(self, filepath: Path | str) -> str:
        # The version is a string, but the user often writes it as an int.
        # YAML will then parse it as an int, for example, `3_0_2` will be parsed as `302`.
        # This is technically a user mistake, as you should quote the version in the YAML file.
        # However, we do not want to put this burden on the user (knowing the intricate workings of YAML),
        # so we fix it here.
        return quote_int_value_by_key_in_yaml(safe_read(filepath, encoding=BUILD_FOLDER_ENCODING), key="version")

    def load_resource_file(
        self, filepath: Path, environment_variables: dict[str, str | None] | None = None
    ) -> list[dict[str, Any]]:
        raw_yaml = load_yaml_inject_variables(
            self.safe_read(filepath),
            environment_variables or {},
            original_filepath=filepath,
        )
        raw_list = raw_yaml if isinstance(raw_yaml, list) else [raw_yaml]

        for item in raw_list:
            model_id = self.get_id(item)
            # Find the GraphQL files adjacent to the DML files
            graphql_file = filepath.with_suffix(".graphql")
            if not graphql_file.is_file():
                raise ToolkitFileNotFoundError(
                    f"Failed to find GraphQL file. Expected {graphql_file.name} adjacent to {filepath.as_posix()}"
                )

            self._graphql_filepath_cache[model_id] = graphql_file
            graphql_content = safe_read(graphql_file, encoding=BUILD_FOLDER_ENCODING)

            sdk_model_id = DataModelId(space=model_id.space, external_id=model_id.external_id, version=model_id.version)
            parser = GraphQLParser(graphql_content, sdk_model_id)
            try:
                for view in parser.get_views():
                    view_ref = ViewReference(
                        space=view.space, external_id=view.external_id, version=str(view.version or "")
                    )
                    self._datamodels_by_view_id[view_ref].add(model_id)
                deps: set[ViewReference | DataModelReference] = set()
                for dep in parser.get_dependencies():
                    if isinstance(dep, DataModelId):
                        deps.add(
                            DataModelReference(space=dep.space, external_id=dep.external_id, version=str(dep.version))
                        )
                    elif isinstance(dep, ViewId):
                        deps.add(
                            ViewReference(space=dep.space, external_id=dep.external_id, version=str(dep.version or ""))
                        )
                self._dependencies_by_datamodel_id[model_id] = deps
            except Exception as e:
                # We catch a broad exception here to give a more user-friendly error message.
                raise GraphQLParseError(f"Failed to parse GraphQL file {graphql_file.as_posix()}: {e}") from e

            # Add hash to description
            description = item.get("description", "")
            hash_ = calculate_hash(graphql_content)[:8]
            suffix = f"{self._hash_name}{hash_}"
            if len(description) + len(suffix) > 1024:
                LowSeverityWarning(f"Description is above limit for {model_id}. Truncating...").print_warning()
                description = description[: 1024 - len(suffix) + 1 - 3] + "..."
            description += f" {suffix}"
            item["description"] = description
            item["graphqlFile"] = hash_
        return raw_list

    def dump_resource(self, resource: GraphQLDataModelResponse, local: dict[str, Any] | None = None) -> dict[str, Any]:
        dumped = resource.as_request_resource().dump()
        local = local or {}
        for key in ["dml", "preserveDml"]:
            # Local values that are not returned from the API
            if key in local:
                dumped[key] = local[key]

        description = resource.description or ""
        if match := re.match(rf"(.|\n)*( {self._hash_name}([a-f0-9]{{8}}))$", description):
            dumped["graphqlFile"] = match.group(3)
        return dumped

    def create(self, items: Sequence[GraphQLDataModelRequest]) -> list[GraphQLDataModelResponse]:
        creation_order = self._topological_sort(items)

        created_list: list[GraphQLDataModelResponse] = []
        for item in creation_order:
            item_id = item.as_id()
            graphql_file_content = self._get_graphql_content(item_id)
            if "--verbose" in sys.argv:
                print(f"Deploying GraphQL schema {item_id}")

            item_with_dml = item.model_copy(update={"dml": graphql_file_content})
            created = self.client.tool.graphql_data_models.create([item_with_dml])
            created_list.extend(created)
        return created_list

    def _get_graphql_content(self, data_model_id: DataModelReference) -> str:
        filepath = self._graphql_filepath_cache.get(data_model_id)
        if filepath is None:
            raise ToolkitFileNotFoundError(f"Could not find the GraphQL file for {data_model_id}")
        return safe_read(filepath)

    def retrieve(self, ids: Sequence[DataModelReference]) -> list[GraphQLDataModelResponse]:
        return self.client.tool.graphql_data_models.retrieve(list(ids), inline_views=False)

    def update(self, items: Sequence[GraphQLDataModelRequest]) -> list[GraphQLDataModelResponse]:
        return self.create(items)

    def delete(self, ids: Sequence[DataModelReference]) -> int:
        retrieved = self.retrieve(ids)
        views = {view for dml in retrieved for view in dml.views or []}
        self.client.tool.graphql_data_models.delete(list(ids))
        deleted = len(ids)
        if views:
            self.client.tool.views.delete(list(views))
            deleted += len(views)
        return deleted

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: Sequence[Hashable] | None = None,
    ) -> Iterable[GraphQLDataModelResponse]:
        filter_ = DataModelFilter(space=space) if space else None
        for batch in self.client.tool.graphql_data_models.iterate(filter=filter_):
            yield from batch

    def count(self, ids: Sequence[DataModelReference]) -> int:
        retrieved = self.retrieve(ids)
        return sum(len(d.views or []) for d in retrieved)

    def drop_data(self, ids: Sequence[DataModelReference]) -> int:
        return self.delete(ids)

    def _topological_sort(self, items: Sequence[GraphQLDataModelRequest]) -> list[GraphQLDataModelRequest]:
        to_sort = {item.as_id(): item for item in items}
        dependencies: dict[DataModelReference, set[DataModelReference]] = {}
        for item in items:
            item_id = item.as_id()
            dependencies[item_id] = set()
            for dependency in self._dependencies_by_datamodel_id.get(item_id, []):
                if isinstance(dependency, DataModelReference) and dependency in to_sort:
                    dependencies[item_id].add(dependency)
                elif isinstance(dependency, ViewReference):
                    for model_id in self._datamodels_by_view_id.get(dependency, set()):
                        if model_id in to_sort:
                            dependencies[item_id].add(model_id)
        try:
            return [to_sort[item_id] for item_id in TopologicalSorter(dependencies).static_order()]
        except CycleError as e:
            raise ToolkitCycleError(
                f"Cannot create GraphQL schemas. Cycle detected between models {e.args} using the @import directive.",
                *e.args[1:],
            )


@final
class EdgeCRUD(ResourceContainerCRUD[TypedEdgeIdentifier, EdgeRequest, EdgeResponse]):
    item_name = "edges"
    folder_name = "data_modeling"
    resource_cls = EdgeResponse
    resource_write_cls = EdgeRequest
    kind = "Edge"
    yaml_cls = EdgeYAML
    dependencies = frozenset({SpaceCRUD, ViewCRUD, ContainerCRUD, NodeCRUD})
    _doc_url = "Instances/operation/applyNodeAndEdges"

    @property
    def display_name(self) -> str:
        return "edges"

    @classmethod
    def get_required_capability(
        cls, items: Sequence[EdgeRequest] | None, read_only: bool
    ) -> Capability | list[Capability]:
        if not items and items is not None:
            return []

        actions = (
            [DataModelInstancesAcl.Action.Read]
            if read_only
            else [DataModelInstancesAcl.Action.Read, DataModelInstancesAcl.Action.Write]
        )

        return DataModelInstancesAcl(
            actions,
            DataModelInstancesAcl.Scope.SpaceID(list({item.space for item in items}))
            if items is not None
            else DataModelInstancesAcl.Scope.All(),
        )

    @classmethod
    def get_id(cls, item: EdgeRequest | EdgeResponse | dict) -> TypedEdgeIdentifier:
        if isinstance(item, dict):
            if missing := tuple(k for k in {"space", "externalId"} if k not in item):
                # We need to raise a KeyError with all missing keys to get the correct error message.
                raise KeyError(*missing)
            return TypedEdgeIdentifier(space=item["space"], external_id=item["externalId"])
        return item.as_id()

    @classmethod
    def dump_id(cls, id: TypedEdgeIdentifier) -> dict[str, Any]:
        return id.dump()

    @classmethod
    def as_str(cls, id: TypedEdgeIdentifier) -> str:
        return sanitize_filename(f"{id.space}_{id.external_id}")

    @classmethod
    def get_dependent_items(cls, item: dict) -> Iterable[tuple[type[ResourceCRUD], Hashable]]:
        if "space" in item:
            yield SpaceCRUD, SpaceReference(space=item["space"])
        for source in item.get("sources", []):
            if (identifier := source.get("source")) and isinstance(identifier, dict):
                if identifier.get("type") == "view" and in_dict(("space", "externalId", "version"), identifier):
                    yield (
                        ViewCRUD,
                        ViewReference(
                            space=identifier["space"],
                            external_id=identifier["externalId"],
                            version=str(v) if (v := identifier.get("version")) else "",
                        ),
                    )
                elif identifier.get("type") == "container" and in_dict(("space", "externalId"), identifier):
                    yield (
                        ContainerCRUD,
                        ContainerReference(space=identifier["space"], external_id=identifier["externalId"]),
                    )

        for key in ["startNode", "endNode", "type"]:
            if node_ref := item.get(key):
                if isinstance(node_ref, dict) and in_dict(("space", "externalId"), node_ref):
                    yield NodeCRUD, TypedNodeIdentifier(space=node_ref["space"], external_id=node_ref["externalId"])

    def dump_resource(self, resource: EdgeResponse, local: dict[str, Any] | None = None) -> dict[str, Any]:
        # CDF resource does not have properties set, so we need to do a lookup
        local = local or {}
        sources = [
            TypedViewReference._load(source["source"]) for source in local.get("sources", []) if "source" in source
        ]

        # Default dump
        dumped = resource.as_request_resource().dump()
        if sources:
            try:
                node_id = resource.as_id()
                res = self.client.tool.instances.retrieve([node_id], source=sources[0])
            except ToolkitAPIError:
                ...
            else:
                if res:
                    # Dump again with properties from the source view.
                    dumped = res[0].as_request_resource().dump()

        if "existingVersion" not in local:
            # Existing version is typically not set when creating nodes, but we get it back
            # when we retrieve the node from the server.
            dumped.pop("existingVersion", None)
        if dumped.get("instanceType") == "edge" and "instanceType" not in local:
            # Toolkit uses file suffix to determine instanceType, so we need to remove it from the CDF resource
            # to match the local resource.
            dumped.pop("instanceType", None)

        return dumped

    def create(self, items: Sequence[EdgeRequest]) -> list[InstanceSlimDefinition]:
        return self.client.tool.instances.create(list(items))

    def retrieve(self, ids: Sequence[TypedEdgeIdentifier]) -> list[EdgeResponse]:
        results = self.client.tool.instances.retrieve(list(ids))
        return [r for r in results if isinstance(r, EdgeResponse)]

    def update(self, items: Sequence[EdgeRequest]) -> list[InstanceSlimDefinition]:
        return self.client.tool.instances.create(list(items))

    def delete(self, ids: Sequence[TypedEdgeIdentifier]) -> int:
        try:
            deleted = self.client.tool.instances.delete(list(ids))
        except ToolkitAPIError as e:
            if "not exist" in str(e) and "space" in str(e).lower():
                return 0
            raise e
        return len(deleted)

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: Sequence[Hashable] | None = None,
    ) -> Iterable[EdgeResponse]:
        filter_ = InstanceFilter(
            instance_type="edge",
            space=[space] if space else None,
        )
        for batch in self.client.tool.instances.iterate(filter=filter_):
            for inst in batch:
                if isinstance(inst, EdgeResponse):
                    yield inst

    def count(self, ids: Sequence[TypedEdgeIdentifier]) -> int:
        return len(ids)

    def drop_data(self, ids: Sequence[TypedEdgeIdentifier]) -> int:
        # Edges will be deleted in .delete call.
        return 0

    def diff_list(
        self, local: list[Any], cdf: list[Any], json_path: tuple[str | int, ...]
    ) -> tuple[dict[int, int], list[int]]:
        if json_path == ("sources",):
            return diff_list_identifiable(local, cdf, get_identifier=lambda x: dm_identifier(x["source"]))
        return super().diff_list(local, cdf, json_path)



from cognite_toolkit._cdf_tk.client._resource_base import (
    BaseModelObject,
    RequestResource,
    ResponseResource,
)

from .identifiers import SequenceRowId
from .sequence import SequenceColumnSlim


class SequenceRow(BaseModelObject):
    """Represents a row in a sequence."""

    row_number: int
    values: list[str | int | float]


class SequenceRows(BaseModelObject):
    """Base class for sequence rows with common fields."""

    rows: list[SequenceRow]


class SequenceRowsRequest(SequenceRows, RequestResource):
    """Request resource for inserting sequence rows."""

    external_id: str
    columns: list[str]

    def as_id(self) -> SequenceRowId:
        return SequenceRowId(external_id=self.external_id, rows=tuple([row.row_number for row in self.rows]))


class SequenceRowsResponse(SequenceRows, ResponseResource[SequenceRowsRequest]):
    """Response resource for sequence rows.

    Note: The CDF API for sequence rows typically doesn't return the standard
    created_time/last_updated_time fields in the same way as other resources.
    The response contains the sequence identification and the row data.
    """

    external_id: str | None = None
    id: int
    columns: list[SequenceColumnSlim]
    next_cursor: str | None = None

    @classmethod
    def request_cls(cls) -> type[SequenceRowsRequest]:
        return SequenceRowsRequest

    def as_request_resource(self) -> SequenceRowsRequest:
        dumped = self.dump()
        dumped["columns"] = [col.external_id for col in self.columns]
        return SequenceRowsRequest.model_validate(dumped, extra="ignore")



from abc import ABC
from typing import Annotated, Any, Literal

from pydantic import Field, TypeAdapter, field_serializer
from pydantic_core.core_schema import FieldSerializationInfo

from cognite_toolkit._cdf_tk.client._resource_base import BaseModelObject

from ._references import ContainerReference


class ConstraintDefinition(BaseModelObject, ABC):
    constraint_type: str


class UniquenessConstraintDefinition(ConstraintDefinition):
    constraint_type: Literal["uniqueness"] = "uniqueness"
    properties: list[str]
    by_space: bool | None = None


class RequiresConstraintDefinition(ConstraintDefinition):
    constraint_type: Literal["requires"] = "requires"
    require: ContainerReference

    @field_serializer("require", mode="plain")
    @classmethod
    def serialize_require(cls, require: ContainerReference, info: FieldSerializationInfo) -> dict[str, Any]:
        return {**require.model_dump(**vars(info)), "type": "container"}


Constraint = Annotated[
    UniquenessConstraintDefinition | RequiresConstraintDefinition,
    Field(discriminator="constraint_type"),
]

ConstraintAdapter: TypeAdapter[Constraint] = TypeAdapter(Constraint)



from abc import ABC

from cognite_toolkit._cdf_tk.client._resource_base import BaseModelObject, RequestResource, ResponseResource

from ._references import SpaceReference


class Space(BaseModelObject, ABC):
    space: str
    name: str | None = None
    description: str | None = None

    def as_id(self) -> SpaceReference:
        return SpaceReference(space=self.space)


class SpaceRequest(Space, RequestResource): ...


class SpaceResponse(Space, ResponseResource[SpaceRequest]):
    created_time: int
    last_updated_time: int
    is_global: bool

    @classmethod
    def request_cls(cls) -> type[SpaceRequest]:
        return SpaceRequest



from typing import Any, Literal, TypeAlias

from pydantic import Field, field_serializer

from cognite_toolkit._cdf_tk.client._resource_base import BaseModelObject, RequestResource, ResponseResource

from .identifiers import ExternalId


class Disabled(BaseModelObject):
    enabled: Literal[False] = False


class ScheduleConfig(BaseModelObject):
    """Schedule configuration for the routine."""

    enabled: Literal[True] = True
    cron_expression: str


class DataSamplingConfig(BaseModelObject):
    """Data sampling configuration for the routine."""

    enabled: Literal[True] = True
    validation_window: int | None = None
    sampling_window: int
    granularity: int


Aggregate = Literal["average", "interpolation", "stepInterpolation"]
LogicalCheckOperator = Literal["eq", "ne", "gt", "ge", "lt", "le"]


class LogicalCheckConfig(BaseModelObject):
    """Logical check configuration for validating time series data."""

    enabled: bool = True
    timeseries_external_id: str
    aggregate: Aggregate
    operator: LogicalCheckOperator
    value: float


class SteadyStateDetectionConfig(BaseModelObject):
    enabled: bool = True
    timeseries_external_id: str | None = None
    aggregate: Aggregate
    min_section_size: int
    var_threshold: float
    slope_threshold: float


ValueType: TypeAlias = Literal["STRING", "DOUBLE", "STRING_ARRAY", "DOUBLE_ARRAY"]


class Unit(BaseModelObject):
    """Unit definition for inputs and outputs."""

    name: str
    quantity: str


class RoutineInputConstantConfig(BaseModelObject):
    """Input configuration for the routine."""

    name: str
    reference_id: str
    value: str | float | list[str] | list[float]
    value_type: ValueType = "STRING"
    unit: Unit | None = None
    save_timeseries_external_id: str | None = None


class RoutineInputTimeseriesConfig(BaseModelObject):
    """Input configuration for the routine."""

    name: str
    reference_id: str
    source_external_id: str
    aggregate: Aggregate
    unit: Unit | None = None
    save_timeseries_external_id: str | None = None


class RoutineOutputConfig(BaseModelObject):
    """Output configuration for the routine."""

    name: str
    reference_id: str
    unit: Unit | None = None
    value_type: ValueType = "STRING"
    save_timeseries_external_id: str | None = None


# ========== Main Configuration ==========
class SimulatorRoutineConfiguration(BaseModelObject):
    """Complete configuration for a simulator routine revision."""

    schedule: ScheduleConfig | Disabled
    data_sampling: DataSamplingConfig | Disabled
    logical_check: list[LogicalCheckConfig] = Field(default_factory=list)
    steady_state_detection: list[SteadyStateDetectionConfig] = Field(default_factory=list)
    inputs: list[RoutineInputConstantConfig] | None = None
    outputs: list[RoutineOutputConfig] | None = None

    @field_serializer("schedule", "data_sampling")
    def ensure_enabled_is_set(self, value: Any) -> Any:
        # Ensure that schedule and data_sampling always serialize "enabled" field
        if isinstance(value, dict):
            if "enabled" not in value:
                value["enabled"] = True
        return value


class ScriptStep(BaseModelObject):
    """A Get or Set step within a script stage."""

    order: int
    description: str | None = None
    step_type: Literal["Get", "Set", "Command"]
    arguments: dict[str, Any]


class ScriptStage(BaseModelObject):
    """A stage in the script containing multiple steps."""

    order: int
    description: str | None = None
    steps: list[ScriptStep]


class SimulatorRoutineRevision(BaseModelObject):
    external_id: str
    routine_external_id: str
    configuration: SimulatorRoutineConfiguration
    script: list[ScriptStage] | None = None


class SimulatorRoutineRevisionRequest(RequestResource, SimulatorRoutineRevision):
    """Request class for creating a simulator routine revision."""

    def as_id(self) -> ExternalId:
        return ExternalId(external_id=self.external_id)


class SimulatorRoutineRevisionResponse(ResponseResource[SimulatorRoutineRevisionRequest], SimulatorRoutineRevision):
    """Response class for a simulator routine revision."""

    id: int
    simulator_external_id: str
    model_external_id: str
    simulator_integration_external_id: str
    created_by_user_id: str
    version_number: int
    data_set_id: int
    created_time: int

    @classmethod
    def request_cls(cls) -> type[SimulatorRoutineRevisionRequest]:
        return SimulatorRoutineRevisionRequest



from abc import abstractmethod

from pydantic import BaseModel
from pydantic.alias_generators import to_camel

from cognite_toolkit._cdf_tk.client._resource_base import Identifier


class BaseModelResource(BaseModel, alias_generator=to_camel, extra="forbid"): ...


class ToolkitResource(BaseModelResource):
    @abstractmethod
    def as_id(self) -> Identifier:
        """Return an identifier for this resource."""
        raise NotImplementedError()



from abc import ABC, abstractmethod
from collections import defaultdict
from collections.abc import Sequence
from threading import Lock
from typing import Literal, TypeAlias

from pydantic import BaseModel
from pydantic.alias_generators import to_camel

from cognite_toolkit._cdf_tk.utils.fileio import NDJsonWriter


class LogEntry(BaseModel, alias_generator=to_camel, extra="ignore", populate_by_name=True):
    """Represents a log entry for tracking storage I/O operations."""

    id: str


OperationStatus: TypeAlias = Literal["success", "failure", "unchanged", "pending"]


class ItemTracker:
    """Tracks issues accumulated for a single item during pipeline processing."""

    def __init__(self, item_id: str) -> None:
        self.item_id = item_id
        self.issues: list[str] = []

    def add_issue(self, issue: str) -> None:
        """Add an issue encountered during processing."""
        self.issues.append(issue)


class OperationTracker(ABC):
    """Abstract base class for operation trackers."""

    @abstractmethod
    def add_issue(self, item_id: str, issue: str) -> None:
        """Add an issue to an item."""
        raise NotImplementedError()

    @abstractmethod
    def finalize_item(self, item_id: str | list[str], status: OperationStatus) -> None:
        """Finalize an item with its final status."""
        raise NotImplementedError()

    @abstractmethod
    def get_status_counts(self) -> dict[OperationStatus, int]:
        """Get counts per final status."""
        raise NotImplementedError()

    @abstractmethod
    def get_issue_counts(self, status: OperationStatus) -> dict[str, int]:
        """Get issue counts, optionally filtered by status."""
        raise NotImplementedError()


class NoOpTracker(OperationTracker):
    """A no-op tracker that does nothing."""

    def add_issue(self, item_id: str, issue: str) -> None:
        """No-op: Discard the issue."""
        pass

    def finalize_item(self, item_id: str | list[str], status: OperationStatus) -> None:
        """No-op: Do nothing."""
        pass

    def get_status_counts(self) -> dict[OperationStatus, int]:
        """Return empty status counts."""
        return {}

    def get_issue_counts(self, status: OperationStatus) -> dict[str, int]:
        """Return empty issue counts."""
        return {}


class MemoryOperationTracker(OperationTracker):
    """Tracks the overall operation progress and issues across multiple items.

    Tracks counts of final statuses and issues per status.
    """

    def __init__(self) -> None:
        self._lock = Lock()
        self._active_items: dict[str, ItemTracker] = {}
        self._status_counts: dict[OperationStatus, int] = defaultdict(int)
        self._issue_counts: dict[OperationStatus, dict[str, int]] = defaultdict(lambda: defaultdict(int))

    def add_issue(self, item_id: str, issue: str) -> None:
        """Add an issue to an item, creating tracker if needed."""
        with self._lock:
            if item_id not in self._active_items:
                self._active_items[item_id] = ItemTracker(item_id)
            self._active_items[item_id].add_issue(issue)

    def finalize_item(self, item_id: str | list[str], status: OperationStatus) -> None:
        """Finalize an item with its final status.

        Args:
            item_id: The item's identifier.
            status: Final status (success, failure, unchanged).
        """
        with self._lock:
            if isinstance(item_id, list):
                for iid in item_id:
                    self._finalize_item_unlocked(iid, status)
            else:
                self._finalize_item_unlocked(item_id, status)

    def _finalize_item_unlocked(self, item_id: str, status: OperationStatus) -> None:
        """Internal method to finalize an item without acquiring the lock."""
        tracker = self._active_items.pop(item_id, None)
        self._status_counts[status] += 1
        if tracker is not None:
            for issue in tracker.issues:
                self._issue_counts[status][issue] += 1

    def get_status_counts(self) -> dict[OperationStatus, int]:
        """Get counts per final status."""
        with self._lock:
            return dict(self._status_counts)

    def get_issue_counts(self, status: OperationStatus) -> dict[str, int]:
        """Get issue counts, optionally filtered by status."""
        with self._lock:
            return dict(self._issue_counts.get(status, {}))


class DataLogger(ABC):
    """Abstract base class for data loggers that track operations and log entries."""

    tracker: OperationTracker

    @abstractmethod
    def log(self, entry: LogEntry | Sequence[LogEntry]) -> None:
        """Log a detailed entry."""
        raise NotImplementedError()


class NoOpLogger(DataLogger):
    """A no-op logger that discards all log entries and does no tracking."""

    def __init__(self) -> None:
        self.tracker = NoOpTracker()

    def log(self, entry: LogEntry | Sequence[LogEntry]) -> None:
        """Discard the log entry (no-op)."""
        pass


class FileDataLogger(DataLogger):
    """Composes aggregation tracking with detailed file logging."""

    def __init__(self, writer: NDJsonWriter) -> None:
        self.tracker = MemoryOperationTracker()
        self._writer = writer

    def log(self, entry: LogEntry | Sequence[LogEntry]) -> None:
        """Log a detailed entry to the file."""
        entries = entry if isinstance(entry, Sequence) else [entry]
        self._writer.write_chunks([e.model_dump(by_alias=True) for e in entries])



from collections import defaultdict
from typing import Any, ClassVar, Literal

from cognite_toolkit._cdf_tk.client._resource_base import (
    BaseModelObject,
    Identifier,
    ResponseResource,
    UpdatableRequestResource,
)
from cognite_toolkit._cdf_tk.client._types import Metadata

from .identifiers import ExternalId


class SequenceColumnSlim(BaseModelObject):
    external_id: str
    name: str | None = None
    value_type: Literal["STRING", "DOUBLE", "LONG"] | None = None


class SequenceColumn(SequenceColumnSlim):
    description: str | None = None
    metadata: Metadata | None = None


class SequenceColumnRequest(SequenceColumn, UpdatableRequestResource):
    def as_id(self) -> Identifier:
        return ExternalId(external_id=self.external_id)


class SequenceColumnResponse(SequenceColumn, ResponseResource[SequenceColumnRequest]):
    created_time: int
    last_updated_time: int

    @classmethod
    def request_cls(cls) -> type[SequenceColumnRequest]:
        return SequenceColumnRequest


class Sequence(BaseModelObject):
    external_id: str | None = None
    name: str | None = None
    description: str | None = None
    asset_id: int | None = None
    data_set_id: int | None = None
    metadata: Metadata | None = None

    def as_id(self) -> ExternalId:
        if self.external_id is None:
            raise ValueError("Cannot convert Sequence to ExternalId when external_id is None")
        return ExternalId(external_id=self.external_id)


class SequenceRequest(Sequence, UpdatableRequestResource):
    container_fields: ClassVar[frozenset[str]] = frozenset({"metadata", "columns"})
    columns: list[SequenceColumnRequest]

    def as_update(
        self, mode: Literal["patch", "replace"], last_columns: list[SequenceColumn] | None = None
    ) -> dict[str, Any]:
        output = super().as_update(mode)
        output["update"].pop("columns", None)
        exclude_unset = mode == "patch"
        existing_ids = {col.external_id for col in last_columns} if last_columns is not None else set()
        columns: dict[str, list[Any]] = defaultdict(list)
        for col in self.columns:
            col_dumped = col.model_dump(mode="json", by_alias=True, exclude_none=True)
            if col_dumped == {"externalId": col.external_id}:
                continue  # No changes to this column

            if col.external_id in existing_ids:
                columns["modify"].append(col.as_update(mode=mode))
            else:
                columns["add"].append(col.model_dump(mode="json", by_alias=True, exclude_unset=exclude_unset))
        if mode == "replace":
            current_columns_ids = {col.external_id for col in self.columns}
            for last_col in last_columns or []:
                if last_col.external_id not in current_columns_ids:
                    columns["remove"].append({"externalId": last_col.external_id})
        if columns:
            output["update"]["columns"] = dict(columns)
        return output


class SequenceResponse(Sequence, ResponseResource[SequenceRequest]):
    id: int
    created_time: int
    last_updated_time: int
    columns: list[SequenceColumnResponse]

    @classmethod
    def request_cls(cls) -> type[SequenceRequest]:
        return SequenceRequest



from collections import UserList, defaultdict

from pydantic import BaseModel


class Insight(BaseModel):
    """Base class for all insights"""

    message: str
    code: str | None = None
    fix: str | None = None

    @classmethod
    def insight_type(cls) -> str:
        return cls.__name__


class ModelSyntaxError(Insight):
    """If any syntax error is found. Stop validation
    and ask user to fix the syntax error first."""

    ...


class ConsistencyError(Insight):
    """If any consistency error is found, the deployment of the CDF resource will fail."""

    ...


class ConsistencyWarning(Insight):
    """Typically handles validations with extras=True, where internal representation might be off sync with
    CDF API definition.
    """

    ...


class Recommendation(Insight):
    """Best practice recommendation."""

    ...


class InsightList(UserList[Insight]):
    """A list of insights that can be sorted by type and message."""

    def by_type(self) -> dict[type[Insight], list[Insight]]:
        """Returns a dictionary of insights sorted by their type."""
        result: dict[type[Insight], list[Insight]] = defaultdict(list)
        for insight in self.data:
            insight_type = type(insight)
            if insight_type not in result:
                result[insight_type] = []
            result[insight_type].append(insight)
        return result

    def by_code(self) -> dict[str, list[Insight]]:
        """Returns a dictionary of insights sorted by their code."""
        result: dict[str, list[Insight]] = defaultdict(list)
        for insight in self.data:
            if insight.code is not None:
                result[insight.code].append(insight)
            else:
                result["UNDEFINED"].append(insight)
        return dict(result)

    @property
    def has_model_syntax_errors(self) -> bool:
        """Returns True if there are any model syntax errors in the insights."""
        return any(isinstance(insight, ModelSyntaxError) for insight in self.data)



from collections.abc import Sequence
from typing import Any

from cognite_toolkit._cdf_tk.client.api.instances import MultiWrappedInstancesAPI, WrappedInstancesAPI
from cognite_toolkit._cdf_tk.client.cdf_client import PagedResponse, ResponseItems
from cognite_toolkit._cdf_tk.client.http_client import (
    HTTPClient,
    ItemsSuccessResponse,
    SuccessResponse,
)
from cognite_toolkit._cdf_tk.client.resource_classes.apm_config_v1 import (
    APM_CONFIG_SPACE,
    APMConfigRequest,
    APMConfigResponse,
)
from cognite_toolkit._cdf_tk.client.resource_classes.data_modeling._query import QueryResponseUntyped
from cognite_toolkit._cdf_tk.client.resource_classes.infield import (
    DataExplorationConfig,
    InFieldCDMLocationConfigRequest,
    InFieldCDMLocationConfigResponse,
    InFieldLocationConfig,
    InFieldLocationConfigRequest,
    InFieldLocationConfigResponse,
)
from cognite_toolkit._cdf_tk.client.resource_classes.instance_api import (
    TypedInstanceIdentifier,
    TypedNodeIdentifier,
)


class InfieldConfigAPI(MultiWrappedInstancesAPI[InFieldLocationConfigRequest, InFieldLocationConfigResponse]):
    _LOCATION_REF = "locationConfig"
    _EXPLORATION_REF = "dataExplorationConfig"

    def __init__(self, http_client: HTTPClient) -> None:
        # 500 is chosen as 1000 is the maximum for nodes, and each location config consists of 1 or 2 nodes
        super().__init__(http_client, query_chunk=500)

    def _retrieve_query(self, items: Sequence[TypedInstanceIdentifier]) -> dict[str, Any]:
        return {
            "with": {
                self._LOCATION_REF: {
                    "limit": len(items),
                    "nodes": {
                        "filter": {
                            "instanceReferences": [
                                {"space": item.space, "externalId": item.external_id} for item in items
                            ]
                        },
                    },
                },
                self._EXPLORATION_REF: {
                    "nodes": {
                        "from": "locationConfig",
                        "direction": "outwards",
                        "through": {
                            "source": InFieldLocationConfig.VIEW_ID.dump(),
                            "identifier": "dataExplorationConfig",
                        },
                    }
                },
            },
            "select": {
                self._LOCATION_REF: {
                    "sources": [{"source": InFieldLocationConfig.VIEW_ID.dump(), "properties": ["*"]}],
                },
                self._EXPLORATION_REF: {
                    "sources": [{"source": DataExplorationConfig.VIEW_ID.dump(), "properties": ["*"]}],
                },
            },
        }

    def _validate_query_response(self, query_response: QueryResponseUntyped) -> list[InFieldLocationConfigResponse]:
        exploration_config_results = (
            DataExplorationConfig.model_validate(item) for item in query_response.items.get(self._EXPLORATION_REF, [])
        )
        exploration_config_map = {(item.space, item.external_id): item for item in exploration_config_results}
        results: list[InFieldLocationConfigResponse] = []
        for item in query_response.items.get(self._LOCATION_REF, []):
            location_config = InFieldLocationConfigResponse.model_validate(item)
            exploration_config = location_config.data_exploration_config
            if exploration_config is not None:
                location_config.data_exploration_config = exploration_config_map.get(
                    (exploration_config.space, exploration_config.external_id), exploration_config
                )
            results.append(location_config)
        return results


class InFieldCDMConfigAPI(WrappedInstancesAPI[TypedNodeIdentifier, InFieldCDMLocationConfigResponse]):
    def __init__(self, http_client: HTTPClient) -> None:
        super().__init__(http_client, InFieldCDMLocationConfigRequest.VIEW_ID)

    def _validate_response(self, response: SuccessResponse) -> ResponseItems[TypedNodeIdentifier]:
        return ResponseItems[TypedNodeIdentifier].model_validate_json(response.body)

    def _validate_page_response(
        self, response: SuccessResponse | ItemsSuccessResponse
    ) -> PagedResponse[InFieldCDMLocationConfigResponse]:
        return PagedResponse[InFieldCDMLocationConfigResponse].model_validate_json(response.body)


class APMConfigAPI(WrappedInstancesAPI[TypedNodeIdentifier, APMConfigResponse]):
    def __init__(self, http_client: HTTPClient) -> None:
        super().__init__(http_client, APMConfigRequest.VIEW_ID)

    def _validate_response(self, response: SuccessResponse) -> ResponseItems[TypedNodeIdentifier]:
        return ResponseItems[TypedNodeIdentifier].model_validate_json(response.body)

    def _validate_page_response(
        self, response: SuccessResponse | ItemsSuccessResponse
    ) -> PagedResponse[APMConfigResponse]:
        return PagedResponse[APMConfigResponse].model_validate_json(response.body)

    def list(self, limit: int | None = 100) -> list[APMConfigResponse]:
        """List all APM configs.

        Args:
            limit: Maximum number of items to return. If None, all items are returned.

        Returns:
            List of APMConfigResponse objects.
        """
        return super()._list_instances(spaces=[APM_CONFIG_SPACE], instance_type="node", limit=limit)


class InfieldAPI:
    def __init__(self, http_client: HTTPClient) -> None:
        self._http_client = http_client
        self.apm_config = APMConfigAPI(http_client)
        self.config = InfieldConfigAPI(http_client)
        self.cdm_config = InFieldCDMConfigAPI(http_client)



from collections.abc import Iterable, Sequence
from typing import Any

from cognite_toolkit._cdf_tk.client.cdf_client import CDFResourceAPI, PagedResponse
from cognite_toolkit._cdf_tk.client.cdf_client.api import Endpoint
from cognite_toolkit._cdf_tk.client.http_client import HTTPClient, ItemsSuccessResponse, RequestMessage, SuccessResponse
from cognite_toolkit._cdf_tk.client.resource_classes.data_modeling._references import DatapointSubscriptionTimeSeriesId
from cognite_toolkit._cdf_tk.client.resource_classes.datapoint_subscription import (
    DatapointSubscriptionRequest,
    DatapointSubscriptionResponse,
    DatapointSubscriptionUpdateRequest,
)
from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import ExternalId


class DatapointSubscriptionsAPI(CDFResourceAPI[DatapointSubscriptionResponse]):
    def __init__(self, http_client: HTTPClient) -> None:
        super().__init__(
            http_client=http_client,
            method_endpoint_map={
                "create": Endpoint(method="POST", path="/timeseries/subscriptions", item_limit=1),
                "retrieve": Endpoint(method="POST", path="/timeseries/subscriptions/byids", item_limit=1),
                "update": Endpoint(method="POST", path="/timeseries/subscriptions/update", item_limit=1),
                "delete": Endpoint(method="POST", path="/timeseries/subscriptions/delete", item_limit=1),
                "list": Endpoint(method="GET", path="/timeseries/subscriptions", item_limit=100),
            },
        )
        self._list_members = Endpoint(method="GET", path="/timeseries/subscriptions/members", item_limit=100)

    def _validate_page_response(
        self, response: SuccessResponse | ItemsSuccessResponse
    ) -> PagedResponse[DatapointSubscriptionResponse]:
        return PagedResponse[DatapointSubscriptionResponse].model_validate_json(response.body)

    def create(self, items: Sequence[DatapointSubscriptionRequest]) -> list[DatapointSubscriptionResponse]:
        """Create datapoint subscriptions in CDF.

        Args:
            items: List of DatapointSubscriptionRequest objects to create.
        Returns:
            List of created DatapointSubscriptionResponse objects.
        """
        return self._request_item_response(items, "create")

    def retrieve(
        self, items: Sequence[ExternalId], ignore_unknown_ids: bool = False
    ) -> list[DatapointSubscriptionResponse]:
        """Retrieve datapoint subscriptions from CDF.

        Args:
            items: List of ExternalId objects to retrieve.
            ignore_unknown_ids: Whether to ignore unknown IDs.
        Returns:
            List of retrieved DatapointSubscriptionResponse objects.
        """
        return self._request_item_response(
            items, method="retrieve", extra_body={"ignoreUnknownIds": ignore_unknown_ids}
        )

    def update(self, items: Sequence[DatapointSubscriptionUpdateRequest]) -> list[DatapointSubscriptionResponse]:
        """Update datapoint subscriptions in CDF.

        Args:
            items: List of DatapointSubscriptionUpdateRequest objects to update.

        Returns:
            List of updated DatapointSubscriptionResponse objects.
        """
        return self._request_item_response(items, "update")

    def delete(self, items: Sequence[ExternalId], ignore_unknown_ids: bool = False) -> None:
        """Delete datapoint subscriptions from CDF.

        Args:
            items: List of ExternalId objects to delete.
            ignore_unknown_ids: Whether to ignore unknown IDs.
        """
        self._request_no_response(items, "delete", extra_body={"ignoreUnknownIds": ignore_unknown_ids})

    def list_members(self, external_id: str, limit: int | None = 100) -> list[DatapointSubscriptionTimeSeriesId]:
        """List time series IDs subscribed to by a datapoint subscription.

        Args:
            external_id: External ID of the datapoint subscription.
            limit: Maximum number of items to return.

        Returns:
            List of time series IDs subscribed to by the datapoint subscription.
        """
        cursor: str | None = None
        total: int = 0
        endpoint = self._list_members
        result: list[DatapointSubscriptionTimeSeriesId] = []
        while cursor is not None or total == 0:
            page_limit = endpoint.item_limit if limit is None else min(limit - total, endpoint.item_limit)
            parameters: dict[str, Any] = {"externalId": external_id, "limit": page_limit}
            if cursor is not None:
                parameters["cursor"] = cursor
            response = self._http_client.request_single_retries(
                RequestMessage(
                    endpoint_url=self._http_client.config.create_api_url(endpoint.path),
                    method=endpoint.method,
                    parameters=parameters,
                )
            ).get_success_or_raise()
            page_response = PagedResponse[DatapointSubscriptionTimeSeriesId].model_validate_json(response.body)
            result.extend(page_response.items)
            total += len(page_response.items)
            if (limit is not None and total >= limit) or not page_response.items:
                break
            cursor = page_response.next_cursor

        return result

    def paginate(self, limit: int = 100, cursor: str | None = None) -> PagedResponse[DatapointSubscriptionResponse]:
        """Paginate over datapoint subscriptions in CDF.

        Args:
            limit: Maximum number of items to return.
            cursor: Cursor for pagination.

        Returns:
            PagedResponse of DatapointSubscriptionResponse objects.
        """
        return self._paginate(cursor=cursor, limit=limit)

    def iterate(self, limit: int | None = 100) -> Iterable[list[DatapointSubscriptionResponse]]:
        """Iterate over all datapoint subscriptions in CDF.

        Args:
            limit: Maximum number of items to return per page.

        Returns:
            Iterable of lists of DatapointSubscriptionResponse objects.
        """
        return self._iterate(limit=limit)

    def list(self, limit: int | None = 100) -> list[DatapointSubscriptionResponse]:
        """List all datapoint subscriptions in CDF.

        Returns:
            List of DatapointSubscriptionResponse objects.
        """
        return self._list(limit=limit)



from pydantic import Field

from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import ExternalId

from .base import ToolkitResource


class SimulatorRoutineYAML(ToolkitResource):
    """Simulator routine YAML resource class.

    Based on: https://api-docs.cognite.com/20230101/tag/Simulator-Routines/operation/create_simulator_routine_simulators_routines_post
    """

    external_id: str = Field(description="External ID of the simulator routine.", min_length=1, max_length=255)
    model_external_id: str = Field(description="External ID of the simulator model.", min_length=1, max_length=255)
    simulator_integration_external_id: str = Field(
        description="External ID of the simulator integration.", min_length=1, max_length=255
    )
    name: str = Field(description="Name of the simulator routine.", min_length=1, max_length=50)
    description: str | None = Field(None, description="Description of the simulator routine.", max_length=500)

    def as_id(self) -> ExternalId:
        return ExternalId(external_id=self.external_id)



from abc import ABC
from typing import Literal

from pydantic import JsonValue

from cognite_toolkit._cdf_tk.client._resource_base import BaseModelObject, RequestResource, ResponseResource

from ._constraints import Constraint
from ._data_types import DataType
from ._indexes import Index
from ._references import ContainerReference


class ContainerPropertyDefinition(BaseModelObject):
    immutable: bool | None = None
    nullable: bool | None = None
    auto_increment: bool | None = None
    default_value: str | int | float | bool | dict[str, JsonValue] | None = None
    description: str | None = None
    name: str | None = None
    type: DataType


class Container(BaseModelObject, ABC):
    space: str
    external_id: str
    name: str | None = None
    description: str | None = None
    used_for: Literal["node", "edge", "all"] | None = None
    properties: dict[str, ContainerPropertyDefinition]
    constraints: dict[str, Constraint] | None = None
    indexes: dict[str, Index] | None = None

    def as_id(self) -> ContainerReference:
        return ContainerReference(
            space=self.space,
            external_id=self.external_id,
        )


class ContainerRequest(Container, RequestResource): ...


class ContainerResponse(Container, ResponseResource[ContainerRequest]):
    created_time: int
    last_updated_time: int
    is_global: bool

    @classmethod
    def request_cls(cls) -> type[ContainerRequest]:
        return ContainerRequest



from collections.abc import Iterable, Sequence

from cognite_toolkit._cdf_tk.client.cdf_client import CDFResourceAPI, Endpoint, PagedResponse, ResponseItems
from cognite_toolkit._cdf_tk.client.http_client import HTTPClient, ItemsSuccessResponse, RequestMessage, SuccessResponse
from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import RawDatabaseId, RawTableId
from cognite_toolkit._cdf_tk.client.resource_classes.raw import (
    RAWDatabaseRequest,
    RAWDatabaseResponse,
    RawProfileResponse,
    RAWTableRequest,
    RAWTableResponse,
)


class RawDatabasesAPI(CDFResourceAPI[RAWDatabaseResponse]):
    """API for managing RAW databases in CDF.

    This API provides methods to create, list, and delete RAW databases.
    """

    def __init__(self, http_client: HTTPClient) -> None:
        super().__init__(
            http_client,
            {
                "create": Endpoint(method="POST", path="/raw/dbs", item_limit=1000),
                "delete": Endpoint(method="POST", path="/raw/dbs/delete", item_limit=1000),
                "list": Endpoint(method="GET", path="/raw/dbs", item_limit=1000),
            },
        )

    def _validate_page_response(
        self, response: SuccessResponse | ItemsSuccessResponse
    ) -> PagedResponse[RAWDatabaseResponse]:
        return PagedResponse[RAWDatabaseResponse].model_validate_json(response.body)

    def _reference_response(self, response: SuccessResponse) -> ResponseItems[RAWDatabaseResponse]:
        return ResponseItems[RAWDatabaseResponse].model_validate_json(response.body)

    def create(self, items: Sequence[RAWDatabaseRequest]) -> list[RAWDatabaseResponse]:
        """Create databases in CDF.

        Args:
            items: List of RAWDatabase objects to create.

        Returns:
            List of created RAWDatabase objects.
        """
        return self._request_item_response(list(items), "create")

    def delete(self, items: Sequence[RawDatabaseId], recursive: bool = False) -> None:
        """Delete databases from CDF.

        Args:
            items: List of RAWDatabase objects to delete.
            recursive: Whether to delete tables within the database recursively.
        """
        self._request_no_response(items, "delete", extra_body={"recursive": recursive})

    def paginate(
        self,
        limit: int = 100,
        cursor: str | None = None,
    ) -> PagedResponse[RAWDatabaseResponse]:
        """Iterate over all databases in CDF.

        Args:
            limit: Maximum number of items to return.
            cursor: Cursor for pagination.

        Returns:
            PagedResponse of RAWDatabase objects.
        """
        return self._paginate(limit=limit, cursor=cursor)

    def iterate(
        self,
        limit: int = 100,
    ) -> Iterable[list[RAWDatabaseResponse]]:
        """Iterate over all databases in CDF.

        Args:
            limit: Maximum number of items to return per page.

        Returns:
            Iterable of lists of RAWDatabase objects.
        """
        return self._iterate(limit=limit)

    def list(self, limit: int | None = None) -> list[RAWDatabaseResponse]:
        """List all databases in CDF.

        Args:
            limit: Maximum number of databases to return. If None, returns all databases.

        Returns:
            List of RAWDatabase objects.
        """
        return self._list(limit=limit)


class RawTablesAPI(CDFResourceAPI[RAWTableResponse]):
    """API for managing RAW tables in CDF.

    This API provides methods to create, list, and delete RAW tables within a database.

    Note: This API requires db_name as a path parameter for all operations,
    so it overrides several base class methods to handle dynamic paths.
    """

    DEFAULT_PROFILE_LIMIT = 1000
    MAX_PROFILE_LIMIT = 1_000_000

    def __init__(self, http_client: HTTPClient) -> None:
        # We pass empty endpoint map since paths are dynamic (depend on db_name)
        super().__init__(
            http_client,
            {
                "create": Endpoint(method="POST", path="/raw/dbs/{dbName}/tables", item_limit=1000),
                "delete": Endpoint(method="POST", path="/raw/dbs/{dbName}/tables/delete", item_limit=1000),
                "list": Endpoint(method="GET", path="/raw/dbs/{dbName}/tables", item_limit=1000),
            },
        )

    def _validate_page_response(
        self, response: SuccessResponse | ItemsSuccessResponse
    ) -> PagedResponse[RAWTableResponse]:
        """Parse a page response. Note: db_name must be injected separately."""
        return PagedResponse[RAWTableResponse].model_validate_json(response.body)

    def _reference_response(self, response: SuccessResponse) -> ResponseItems[RAWTableResponse]:
        """Parse a reference response. Note: db_name must be injected separately."""
        return ResponseItems[RAWTableResponse].model_validate_json(response.body)

    def create(self, items: Sequence[RAWTableRequest], ensure_parent: bool = False) -> list[RAWTableResponse]:
        """Create tables in a database in CDF.

        Args:
            items: List of RAWTable objects to create.
            ensure_parent: Create database if it doesn't exist already.

        Returns:
            List of created RAWTable objects.
        """
        result: list[RAWTableResponse] = []
        for (db_name,), group in self._group_items_by_text_field(items, "db_name").items():
            if not db_name:
                raise ValueError("db_name must be set on all RAWTable items for creation.")
            endpoint = f"/raw/dbs/{db_name}/tables"
            created = self._request_item_response(
                group, "create", params={"ensureParent": ensure_parent}, endpoint=endpoint
            )
            for table in created:
                table.db_name = db_name
                result.append(table)
        return result

    def delete(self, items: Sequence[RawTableId]) -> None:
        """Delete tables from a database in CDF.

        Args:
            items: List of RAWTable objects to delete.
        """
        for (db_name,), group in self._group_items_by_text_field(items, "db_name").items():
            if not db_name:
                raise ValueError("db_name must be set on all RAWTable items for deletion.")
            endpoint = f"/raw/dbs/{db_name}/tables/delete"
            self._request_no_response(list(group), "delete", endpoint=endpoint)

    def profile(
        self, table: RawTableId, limit: int = DEFAULT_PROFILE_LIMIT, timeout_seconds: int | None = None
    ) -> RawProfileResponse:
        """Profiles a table in the specified database and returns the results.

        This is a hidden endpoint that is not part of the official CDF API. However, it is used by the Fusion UI
        to profile tables in the database. This is implemented internally in Cognite Toolkit as Toolkit offers
        profiling of raw tables. This is used to show how data flows into CDF resources.

        Args:
            table (RawTableId): The identifier of the table to profile.
            limit (int, optional): The maximum number of rows to profile. Defaults to DEFAULT_PROFILE_LIMIT.
            timeout_seconds (int, optional): The timeout for the profiling operation in seconds. Defaults to global_config.timeout_seconds.

        Returns:
            RawProfileResponse: The results of the profiling operation.

        """
        if limit <= 0 or limit > self.MAX_PROFILE_LIMIT:
            raise ValueError(f"Limit must be between 1 and {self.MAX_PROFILE_LIMIT}, got {limit}.")
        response = self._http_client.request_single_retries(
            RequestMessage(
                endpoint_url=self._http_client.config.create_api_url("/profiler/raw"),
                method="POST",
                body_content={"database": table.db_name, "table": table.name, "limit": limit},
                client_timeout=timeout_seconds,
            )
        ).get_success_or_raise()
        return RawProfileResponse.model_validate_json(response.body)

    def paginate(
        self,
        db_name: str,
        limit: int = 100,
        cursor: str | None = None,
    ) -> PagedResponse[RAWTableResponse]:
        """Iterate over all tables in a database in CDF.

        Args:
            db_name: The name of the database to list tables from.
            limit: Maximum number of items to return.
            cursor: Cursor for pagination.

        Returns:
            PagedResponse of RAWTable objects.
        """
        page = self._paginate(cursor=cursor, limit=limit, endpoint_path=f"/raw/dbs/{db_name}/tables")
        for table in page.items:
            table.db_name = db_name
        return page

    def iterate(
        self,
        db_name: str,
        limit: int | None = 100,
    ) -> Iterable[list[RAWTableResponse]]:
        """Iterate over all tables in a database in CDF.

        Args:
            db_name: The name of the database to list tables from.
            limit: Maximum number of items to return per page.

        Returns:
            Iterable of lists of RAWTable objects.
        """
        for table in self._iterate(limit=limit, endpoint_path=f"/raw/dbs/{db_name}/tables"):
            for t in table:
                t.db_name = db_name
            yield table

    def list(self, db_name: str, limit: int | None = None) -> list[RAWTableResponse]:
        """List all tables in a database in CDF.

        Args:
            db_name: The name of the database to list tables from.
            limit: Maximum number of tables to return. If None, returns all tables.

        Returns:
            List of RAWTable objects.
        """
        listed = self._list(limit, endpoint_path=f"/raw/dbs/{db_name}/tables")
        for table in listed:
            table.db_name = db_name
        return listed


class RawAPI:
    def __init__(self, http_client: HTTPClient) -> None:
        self.databases = RawDatabasesAPI(http_client)
        self.tables = RawTablesAPI(http_client)



from typing import Literal

from pydantic import Field, JsonValue, field_serializer

from cognite_toolkit._cdf_tk.client._resource_base import BaseModelObject

from ._instance import InstanceResponse
from ._references import ViewReference


class QuerySortSpec(BaseModelObject):
    """Sorting specification for query results."""

    property: list[str]
    direction: Literal["ascending", "descending"] | None = None
    nulls_first: bool | None = None


class QueryUnitReference(BaseModelObject):
    """Reference to a unit for unit conversion in query results."""

    external_id: str


class QueryUnitSystemReference(BaseModelObject):
    unit_system_name: str


class QueryTargetUnit(BaseModelObject):
    """Target unit conversion specification for a property."""

    property: str
    unit: QueryUnitReference | QueryUnitSystemReference


class QueryThrough(BaseModelObject):
    """Traversal specification through a direct relation property on a view."""

    source: ViewReference
    identifier: str


class QueryExpression(BaseModelObject):
    sort: list[QuerySortSpec] | None = None
    limit: int | None = None


class QueryTableExpression(BaseModelObject):
    from_: str | None = Field(None, alias="from")
    chain_to: Literal["source", "destination"] | None = None
    direction: Literal["outwards", "inwards"] | None = None


class QueryNodeTableExpression(QueryTableExpression):
    """Table expression defining how to select nodes in a query result set."""

    through: QueryThrough | None = None
    filter: dict[str, JsonValue] | None = None


class QueryEdgeTableExpression(QueryTableExpression):
    """Table expression defining how to select edges in a query result set."""

    max_distance: int | None = None
    filter: dict[str, JsonValue] | None = None
    node_filter: dict[str, JsonValue] | None = None
    termination_filter: dict[str, JsonValue] | None = None
    limit_each: int | None = None


class QueryNodeExpression(QueryExpression):
    """A named result set expression in the 'with' clause of a query.

    Must specify exactly one of ``nodes`` or ``edges``.
    """

    nodes: QueryNodeTableExpression


class QueryEdgeExpression(QueryExpression):
    """A named result set expression in the 'with' clause of a query."""

    post_sort: list[QuerySortSpec] | None = None
    edges: QueryEdgeTableExpression


class QuerySelectSource(BaseModelObject):
    """Source selector specifying which view properties to return."""

    source: ViewReference
    properties: list[str]
    target_units: list[QueryTargetUnit] | None = None

    @field_serializer("source")
    def include_type(self, source: ViewReference) -> dict[str, str]:
        return {**source.dump(), "type": "view"}


class QuerySelect(BaseModelObject):
    """Select clause for a result set expression, specifying which properties to return."""

    sources: list[QuerySelectSource] | None = None
    sort: list[QuerySortSpec] | None = None
    limit: int | None = None


class QueryDebugParameters(BaseModelObject):
    """Debug parameters for the query endpoint."""

    emit_results: bool | None = None
    timeout: int | None = None
    profile: bool | None = None


class QueryRequest(BaseModelObject):
    """Request body for the ``POST /models/instances/query`` endpoint.

    See `API docs <https://api-docs.cognite.com/20230101/tag/Instances/operation/queryContent>`_.
    """

    with_: dict[str, QueryNodeExpression | QueryEdgeExpression] = Field(alias="with")
    cursors: dict[str, str] | None = None
    select: dict[str, QuerySelect]
    parameters: dict[str, JsonValue] | None = None
    include_typing: bool | None = None
    debug: QueryDebugParameters | None = None


class QueryResponse(BaseModelObject):
    next_cursor: dict[str, str | None]

    # For now we do not care about the typing and debug structures in the response.
    typing: dict[str, JsonValue] | None = None
    debug: dict[str, JsonValue] | None = None


class QueryResponseTyped(QueryResponse):
    """Response from the ``POST /models/instances/query`` endpoint."""

    items: dict[str, list[InstanceResponse]]


class QueryResponseUntyped(QueryResponse):
    """Response from the ``POST /models/instances/query`` endpoint, without parsing the items into typed instances.

    This is useful when you want to
    """

    items: dict[str, list[dict[str, JsonValue]]]



from collections.abc import Iterable, Sequence
from typing import Literal

from cognite_toolkit._cdf_tk.client.cdf_client import CDFResourceAPI, PagedResponse, ResponseItems
from cognite_toolkit._cdf_tk.client.cdf_client.api import Endpoint
from cognite_toolkit._cdf_tk.client.http_client import HTTPClient, ItemsSuccessResponse, SuccessResponse
from cognite_toolkit._cdf_tk.client.request_classes.filters import ClassicFilter
from cognite_toolkit._cdf_tk.client.resource_classes.asset import AssetRequest, AssetResponse
from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import InternalOrExternalId


class AssetsAPI(CDFResourceAPI[AssetResponse]):
    def __init__(self, http_client: HTTPClient) -> None:
        super().__init__(
            http_client=http_client,
            method_endpoint_map={
                "create": Endpoint(method="POST", path="/assets", item_limit=1000, concurrency_max_workers=1),
                "retrieve": Endpoint(method="POST", path="/assets/byids", item_limit=1000, concurrency_max_workers=1),
                "update": Endpoint(method="POST", path="/assets/update", item_limit=1000, concurrency_max_workers=1),
                "delete": Endpoint(method="POST", path="/assets/delete", item_limit=1000, concurrency_max_workers=1),
                "list": Endpoint(method="POST", path="/assets/list", item_limit=1000),
            },
        )

    def _validate_page_response(self, response: SuccessResponse | ItemsSuccessResponse) -> PagedResponse[AssetResponse]:
        return PagedResponse[AssetResponse].model_validate_json(response.body)

    def _reference_response(self, response: SuccessResponse) -> ResponseItems[InternalOrExternalId]:
        return ResponseItems[InternalOrExternalId].model_validate_json(response.body)

    def create(self, items: Sequence[AssetRequest]) -> list[AssetResponse]:
        """Create assets in CDF.

        Args:
            items: List of AssetRequest objects to create.
        Returns:
            List of created AssetResponse objects.
        """
        return self._request_item_response(items, "create")

    def retrieve(self, items: Sequence[InternalOrExternalId], ignore_unknown_ids: bool = False) -> list[AssetResponse]:
        """Retrieve assets from CDF.

        Args:
            items: List of InternalOrExternalId objects to retrieve.
            ignore_unknown_ids: Whether to ignore unknown IDs.
        Returns:
            List of retrieved AssetResponse objects.
        """
        return self._request_item_response(
            items, method="retrieve", extra_body={"ignoreUnknownIds": ignore_unknown_ids}
        )

    def update(
        self, items: Sequence[AssetRequest], mode: Literal["patch", "replace"] = "replace"
    ) -> list[AssetResponse]:
        """Update assets in CDF.

        Args:
            items: List of AssetRequest objects to update.
            mode: Update mode, either "patch" or "replace".

        Returns:
            List of updated AssetResponse objects.
        """
        return self._update(items, mode=mode)

    def delete(
        self, items: Sequence[InternalOrExternalId], recursive: bool = False, ignore_unknown_ids: bool = False
    ) -> None:
        """Delete assets from CDF.

        Args:
            items: List of InternalOrExternalId objects to delete.
            recursive: Whether to delete assets recursively.
            ignore_unknown_ids: Whether to ignore unknown IDs.
        """
        self._request_no_response(
            items, "delete", extra_body={"recursive": recursive, "ignoreUnknownIds": ignore_unknown_ids}
        )

    def paginate(
        self,
        aggregated_properties: bool = False,
        filter: ClassicFilter | None = None,
        limit: int = 100,
        cursor: str | None = None,
    ) -> PagedResponse[AssetResponse]:
        """Iterate over all assets in CDF.

        Returns:
            PagedResponse of AssetResponse objects.
        """
        return self._paginate(
            cursor=cursor,
            limit=limit,
            body={
                "aggregatedProperties": ["childCount", "path", "depth"] if aggregated_properties else [],
                "filter": filter.dump() if filter else None,
            },
        )

    def iterate(
        self,
        aggregated_properties: bool = False,
        filter: ClassicFilter | None = None,
        limit: int | None = 100,
    ) -> Iterable[list[AssetResponse]]:
        """Iterate over all assets in CDF.

        Returns:
            Sequence of AssetResponse objects.
        """
        return self._iterate(
            limit=limit,
            body={
                "aggregatedProperties": ["childCount", "path", "depth"] if aggregated_properties else [],
                "filter": filter.dump() if filter else None,
            },
        )

    def list(self, limit: int | None = 100) -> list[AssetResponse]:
        """List all asset references in CDF.

        Returns:
            List of AssetResponse objects.
        """
        return self._list(limit=limit)



# Copyright 2023 Cognite AS
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from collections.abc import Hashable, Iterable, Sequence
from datetime import date, datetime
from typing import Any, final

from cognite.client.data_classes.capabilities import (
    Capability,
    DataModelInstancesAcl,
    FilesAcl,
)
from cognite.client.utils._time import convert_data_modelling_timestamp

from cognite_toolkit._cdf_tk.client.request_classes.filters import ClassicFilter
from cognite_toolkit._cdf_tk.client.resource_classes.cognite_file import CogniteFileRequest, CogniteFileResponse
from cognite_toolkit._cdf_tk.client.resource_classes.data_modeling import NodeReference, SpaceReference
from cognite_toolkit._cdf_tk.client.resource_classes.data_modeling._instance import InstanceSlimDefinition
from cognite_toolkit._cdf_tk.client.resource_classes.filemetadata import FileMetadataRequest, FileMetadataResponse
from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import ExternalId, InternalOrExternalId, NameId
from cognite_toolkit._cdf_tk.client.resource_classes.instance_api import TypedNodeIdentifier
from cognite_toolkit._cdf_tk.cruds._base_cruds import ResourceContainerCRUD, ResourceCRUD
from cognite_toolkit._cdf_tk.exceptions import (
    ToolkitRequiredValueError,
)
from cognite_toolkit._cdf_tk.resource_classes import CogniteFileYAML, FileMetadataYAML
from cognite_toolkit._cdf_tk.utils import (
    in_dict,
)
from cognite_toolkit._cdf_tk.utils.diff_list import diff_list_hashable, diff_list_identifiable, dm_identifier

from .auth import GroupAllScopedCRUD, SecurityCategoryCRUD
from .classic import AssetCRUD
from .data_organization import DataSetsCRUD, LabelCRUD
from .datamodel import NodeCRUD, SpaceCRUD, ViewCRUD


@final
class FileMetadataCRUD(ResourceContainerCRUD[ExternalId, FileMetadataRequest, FileMetadataResponse]):
    item_name = "file contents"
    folder_name = "files"
    resource_cls = FileMetadataResponse
    resource_write_cls = FileMetadataRequest
    yaml_cls = FileMetadataYAML
    kind = "FileMetadata"
    dependencies = frozenset({DataSetsCRUD, GroupAllScopedCRUD, LabelCRUD, AssetCRUD})

    _doc_url = "Files/operation/initFileUpload"

    @property
    def display_name(self) -> str:
        return "file metadata"

    @classmethod
    def get_required_capability(
        cls, items: Sequence[FileMetadataRequest] | None, read_only: bool
    ) -> Capability | list[Capability]:
        if not items and items is not None:
            return []

        actions = [FilesAcl.Action.Read] if read_only else [FilesAcl.Action.Read, FilesAcl.Action.Write]

        scope: FilesAcl.Scope.All | FilesAcl.Scope.DataSet = FilesAcl.Scope.All()  # type: ignore[valid-type]
        if items:
            if data_set_ids := {item.data_set_id for item in items if item.data_set_id}:
                scope = FilesAcl.Scope.DataSet(list(data_set_ids))

        return FilesAcl(actions, scope)

    @classmethod
    def get_id(cls, item: FileMetadataRequest | FileMetadataResponse | dict) -> ExternalId:
        if isinstance(item, dict):
            return ExternalId(external_id=item["externalId"])
        if item.external_id is None:
            raise ToolkitRequiredValueError("FileMetadata must have external_id set.")
        return ExternalId(external_id=item.external_id)

    @classmethod
    def get_internal_id(cls, item: FileMetadataResponse | dict) -> int:
        if isinstance(item, dict):
            return item["id"]
        return item.id

    @classmethod
    def dump_id(cls, id: ExternalId) -> dict[str, Any]:
        return id.dump()

    @classmethod
    def get_dependent_items(cls, item: dict) -> Iterable[tuple[type[ResourceCRUD], Hashable]]:
        if "dataSetExternalId" in item:
            yield DataSetsCRUD, ExternalId(external_id=item["dataSetExternalId"])
        if "securityCategoryNames" in item:
            for security_category in item["securityCategoryNames"]:
                yield SecurityCategoryCRUD, NameId(name=security_category)
        if "labels" in item:
            for label in item["labels"]:
                if isinstance(label, dict):
                    yield LabelCRUD, ExternalId(external_id=label["externalId"])
                elif isinstance(label, str):
                    yield LabelCRUD, ExternalId(external_id=label)
        for asset_external_id in item.get("assetExternalIds", []):
            yield AssetCRUD, ExternalId(external_id=asset_external_id)

    def load_resource(self, resource: dict[str, Any], is_dry_run: bool = False) -> FileMetadataRequest:
        if ds_external_id := resource.pop("dataSetExternalId", None):
            resource["dataSetId"] = self.client.lookup.data_sets.id(ds_external_id, is_dry_run)
        if security_categories_names := resource.pop("securityCategoryNames", []):
            resource["securityCategories"] = self.client.lookup.security_categories.id(
                security_categories_names, is_dry_run
            )
        if asset_external_ids := resource.pop("assetExternalIds", None):
            resource["assetIds"] = self.client.lookup.assets.id(asset_external_ids, is_dry_run)
        return FileMetadataRequest.model_validate(resource)

    def dump_resource(self, resource: FileMetadataResponse, local: dict[str, Any] | None = None) -> dict[str, Any]:
        dumped = resource.as_request_resource().dump()
        if data_set_id := dumped.pop("dataSetId", None):
            dumped["dataSetExternalId"] = self.client.lookup.data_sets.external_id(data_set_id)
        if security_categories := dumped.pop("securityCategories", []):
            dumped["securityCategoryNames"] = self.client.lookup.security_categories.external_id(security_categories)
        if asset_ids := dumped.pop("assetIds", []):
            dumped["assetExternalIds"] = self.client.lookup.assets.external_id(asset_ids)
        return dumped

    def create(self, items: Sequence[FileMetadataRequest]) -> list[FileMetadataResponse]:
        return self.client.tool.filemetadata.create(items, overwrite=True)

    def retrieve(self, ids: Sequence[ExternalId]) -> list[FileMetadataResponse]:
        return self.client.tool.filemetadata.retrieve(list(ids), ignore_unknown_ids=True)

    def update(self, items: Sequence[FileMetadataRequest]) -> list[FileMetadataResponse]:
        return self.client.tool.filemetadata.update(items, mode="replace")

    def delete(self, ids: Sequence[InternalOrExternalId]) -> int:
        if not ids:
            return 0
        self.client.tool.filemetadata.delete(list(ids), ignore_unknown_ids=True)
        return len(ids)

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: Sequence[Hashable] | None = None,
    ) -> Iterable[FileMetadataResponse]:
        filter_ = ClassicFilter.from_asset_subtree_and_data_sets(data_set_id=data_set_external_id)
        for files in self.client.tool.filemetadata.iterate(filter=filter_, limit=None):
            yield from files

    def count(self, ids: Sequence[ExternalId]) -> int:
        return sum(
            1 for meta in self.client.tool.filemetadata.retrieve(list(ids), ignore_unknown_ids=True) if meta.uploaded
        )

    def drop_data(self, ids: Sequence[ExternalId]) -> int:
        existing = self.client.tool.filemetadata.retrieve(list(ids), ignore_unknown_ids=True)
        # File and FileMetadata is tightly coupled, so we need to delete the metadata and recreate it
        # without the source set to delete the file.
        deleted_files = self.delete([meta.as_id() for meta in existing])
        self.create([meta.as_request_resource() for meta in existing])
        return deleted_files


@final
class CogniteFileCRUD(ResourceContainerCRUD[NodeReference, CogniteFileRequest, CogniteFileResponse]):
    template_pattern = "$FILENAME"
    item_name = "file contents"
    folder_name = "files"
    kind = "CogniteFile"
    resource_cls = CogniteFileResponse
    resource_write_cls = CogniteFileRequest
    yaml_cls = CogniteFileYAML
    dependencies = frozenset({GroupAllScopedCRUD, SpaceCRUD, ViewCRUD})

    _doc_url = "Files/operation/initFileUpload"

    @property
    def display_name(self) -> str:
        return "cognite files"

    @classmethod
    def get_id(cls, item: CogniteFileResponse | CogniteFileRequest | dict) -> NodeReference:
        if isinstance(item, dict):
            if missing := tuple(k for k in {"space", "externalId"} if k not in item):
                # We need to raise a KeyError with all missing keys to get the correct error message.
                raise KeyError(*missing)
            return NodeReference(space=item["space"], external_id=item["externalId"])
        return NodeReference(space=item.space, external_id=item.external_id)

    @classmethod
    def dump_id(cls, id: NodeReference) -> dict[str, Any]:
        return id.dump()

    @classmethod
    def get_required_capability(cls, items: Sequence[CogniteFileRequest] | None, read_only: bool) -> list[Capability]:
        if not items and items is not None:
            return []

        file_actions = [FilesAcl.Action.Read] if read_only else [FilesAcl.Action.Read, FilesAcl.Action.Write]
        instance_actions = (
            [DataModelInstancesAcl.Action.Read]
            if read_only
            else [DataModelInstancesAcl.Action.Read, DataModelInstancesAcl.Action.Write]
        )

        scope: DataModelInstancesAcl.Scope.All | DataModelInstancesAcl.Scope.SpaceID = DataModelInstancesAcl.Scope.All()  # type: ignore[valid-type]
        if items:
            if spaces := {item.space for item in items}:
                scope = DataModelInstancesAcl.Scope.SpaceID(list(spaces))
        return [
            FilesAcl(file_actions, FilesAcl.Scope.All()),
            DataModelInstancesAcl(instance_actions, scope),
        ]

    def dump_resource(self, resource: CogniteFileResponse, local: dict[str, Any] | None = None) -> dict[str, Any]:
        dumped = resource.as_request_resource().dump(context="toolkit")
        local = local or {}
        if "existingVersion" not in local:
            # Existing version is typically not set when creating nodes, but we get it back
            # when we retrieve the node from the server.
            dumped.pop("existingVersion", None)
        for key in list(dumped.keys()):
            value = dumped[key]
            if key not in local:
                if value is None or value in ([], {}):
                    dumped.pop(key)
                continue
            local_value = local[key]
            if isinstance(local_value, datetime) and isinstance(value, str):
                dumped[key] = convert_data_modelling_timestamp(value)
            elif isinstance(local_value, date) and isinstance(value, str):
                dumped[key] = date.fromisoformat(value)

        if dumped.get("instanceType") == "node" and "instanceType" not in local:
            dumped.pop("instanceType")
        return dumped

    def diff_list(
        self, local: list[Any], cdf: list[Any], json_path: tuple[str | int, ...]
    ) -> tuple[dict[int, int], list[int]]:
        if json_path == ("tags",):
            return diff_list_hashable(local, cdf)
        elif json_path[0] in ("assets", "category"):
            return diff_list_identifiable(local, cdf, get_identifier=dm_identifier)
        return super().diff_list(local, cdf, json_path)

    def create(self, items: Sequence[CogniteFileRequest]) -> list[InstanceSlimDefinition]:
        return self.client.tool.cognite_files.create(items)

    def retrieve(self, ids: Sequence[NodeReference]) -> list[CogniteFileResponse]:
        return self.client.tool.cognite_files.retrieve(
            [TypedNodeIdentifier(space=id_.space, external_id=id_.external_id) for id_ in ids]
        )

    def update(self, items: Sequence[CogniteFileRequest]) -> list[InstanceSlimDefinition]:
        return self.client.tool.cognite_files.create(items)

    def delete(self, ids: Sequence[NodeReference]) -> int:
        return len(
            self.client.tool.cognite_files.delete(
                [TypedNodeIdentifier(space=id_.space, external_id=id_.external_id) for id_ in ids]
            )
        )

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: Sequence[Hashable] | None = None,
    ) -> Iterable[CogniteFileResponse]:
        # We do not have a way to know the source of the file, so we cannot filter on that.
        return []

    def count(self, ids: Sequence[NodeReference]) -> int:
        return sum(bool(file.is_uploaded or False) for file in self.retrieve(ids))

    def drop_data(self, ids: Sequence[NodeReference]) -> int:
        # Deleting and recreating the file, will remove the file contents but keep the metadata.
        retrieved = self.retrieve(ids)
        self.delete(ids)
        return len(self.create([file.as_request_resource() for file in retrieved]))

    @classmethod
    def get_dependent_items(cls, item: dict) -> Iterable[tuple[type[ResourceCRUD], Hashable]]:
        """Returns all items that this item requires.

        For example, a TimeSeries requires a DataSet, so this method would return the
        DatasetLoader and identifier of that dataset.
        """
        if "space" in item:
            yield SpaceCRUD, SpaceReference(space=item["space"])
        for key in ["source", "category", "type"]:
            if key in item and in_dict(("space", "externalId"), item[key]):
                yield NodeCRUD, TypedNodeIdentifier(space=item[key]["space"], external_id=item[key]["externalId"])
        if "assets" in item:
            for asset in item["assets"]:
                if isinstance(asset, dict) and in_dict(("space", "externalId"), asset):
                    yield NodeCRUD, TypedNodeIdentifier(space=asset["space"], external_id=asset["externalId"])



from typing import ClassVar

from pydantic import JsonValue

from cognite_toolkit._cdf_tk.client._resource_base import (
    BaseModelObject,
    ResponseResource,
    UpdatableRequestResource,
)
from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import ExternalId


class RobotDataPostProcessing(BaseModelObject):
    """DataPostprocessing define types of data processing on data captured by the robot.
    DataPostprocessing enables you to automatically process data captured by the robot.

    Args:
        name: DataPostProcessing name.
        external_id: DataPostProcessing external id. Must be unique for the resource type.
        method: DataPostProcessing method. The method is used to call the right functionality on the robot.
        input_schema: Schema that defines what inputs are needed for the data postprocessing.
            The input are values that configure the data postprocessing, e.g max and min values for a gauge.
        description: Description of DataPostProcessing. Textual description of the DataPostProcessing.
    """

    external_id: str
    name: str
    method: str
    description: str | None = None

    def as_id(self) -> ExternalId:
        return ExternalId(external_id=self.external_id)


class RobotDataPostProcessingRequest(RobotDataPostProcessing, UpdatableRequestResource):
    """Request resource for creating or updating a DataPostProcessing."""

    non_nullable_fields: ClassVar[frozenset[str]] = frozenset({"input_schema"})
    input_schema: dict[str, JsonValue] | None = None


class RobotDataPostProcessingResponse(RobotDataPostProcessing, ResponseResource[RobotDataPostProcessingRequest]):
    """Response resource for a DataPostProcessing."""

    # The response always has input_schema
    input_schema: dict[str, JsonValue]

    @classmethod
    def request_cls(cls) -> type[RobotDataPostProcessingRequest]:
        return RobotDataPostProcessingRequest



from typing import Annotated, Any, ClassVar, Literal

from pydantic import Field

from cognite_toolkit._cdf_tk.client._resource_base import (
    BaseModelObject,
    ResponseResource,
    UpdatableRequestResource,
)
from cognite_toolkit._cdf_tk.constants import SPACE_FORMAT_PATTERN

from .identifiers import DataProductVersionId, SemanticVersion

SpaceId = Annotated[str, Field(pattern=SPACE_FORMAT_PATTERN, max_length=43)]


class ViewInstanceSpaces(BaseModelObject):
    read: list[SpaceId] = Field(default_factory=list)
    write: list[SpaceId] = Field(default_factory=list)


class DataProductVersionView(BaseModelObject):
    external_id: str
    instance_spaces: ViewInstanceSpaces = Field(default_factory=ViewInstanceSpaces)


class DataProductVersionDataModel(BaseModelObject):
    external_id: str
    version: str
    views: list[DataProductVersionView] = Field(default_factory=list)


class DataProductVersionTerms(BaseModelObject):
    usage: str | None = None
    limitations: str | None = None


class DataProductVersion(BaseModelObject):
    data_product_external_id: str = Field(exclude=True)
    version: SemanticVersion
    data_model: DataProductVersionDataModel
    status: Literal["draft", "published", "deprecated"] = "draft"
    description: str | None = None
    terms: DataProductVersionTerms | None = None

    def as_id(self) -> DataProductVersionId:
        return DataProductVersionId(
            data_product_external_id=self.data_product_external_id,
            version=self.version,
        )


class DataProductVersionRequest(DataProductVersion, UpdatableRequestResource):
    container_fields: ClassVar[frozenset[str]] = frozenset()

    def as_update(self, mode: Literal["patch", "replace"]) -> dict[str, Any]:
        # The versions update API uses nested {set}/{setNull}/{modify} operators
        # instead of a flat body, so we must build the payload manually.
        update_item: dict[str, Any] = {"version": self.version}
        update: dict[str, Any] = {}
        exclude_unset = mode == "patch"
        dumped = self.model_dump(mode="json", by_alias=True, exclude_unset=exclude_unset)

        for key in ("status", "description"):
            if key not in dumped:
                continue
            if dumped[key] is None:
                update[key] = {"setNull": True}
            else:
                update[key] = {"set": dumped[key]}

        if "terms" in dumped and dumped["terms"] is not None:
            terms_modify: dict[str, Any] = {}
            for sub_key in ("usage", "limitations"):
                if sub_key in dumped["terms"]:
                    val = dumped["terms"][sub_key]
                    terms_modify[sub_key] = {"setNull": True} if val is None else {"set": val}
            if terms_modify:
                update["terms"] = {"modify": terms_modify}

        if "dataModel" in dumped and dumped["dataModel"] is not None:
            views = dumped["dataModel"].get("views")
            if views is not None:
                update["dataModel"] = {"modify": {"views": {"set": views}}}

        update_item["update"] = update
        return update_item


class DataProductVersionResponse(DataProductVersion, ResponseResource[DataProductVersionRequest]):
    data_product_external_id: str = Field(default="", exclude=True)
    created_time: int
    last_updated_time: int

    @classmethod
    def request_cls(cls) -> type[DataProductVersionRequest]:
        return DataProductVersionRequest

    def as_request_resource(self) -> DataProductVersionRequest:
        data = self.dump(camel_case=False)
        data["data_product_external_id"] = self.data_product_external_id
        return DataProductVersionRequest.model_validate(data, extra="ignore")



from cognite_toolkit._cdf_tk.commands.build_v2.data_classes._insights import Recommendation
from cognite_toolkit._cdf_tk.cruds._resource_cruds.datamodel import DataModelCRUD
from cognite_toolkit._cdf_tk.resource_classes.data_model import DataModelYAML
from cognite_toolkit._cdf_tk.rules._base import ToolkitResourceRule


class DummyDataModelRule(ToolkitResourceRule[DataModelYAML]):
    """This is a dummy rule for demonstration purposes.
    It does not perform any actual validation but serves as a template for implementing real rules.

    ## What it does

    ## Why is this bad?

    ## Example

    """

    code = "DUMMY_MODEL_RULE"
    resource_type = DataModelCRUD.kind
    insight_type = Recommendation

    def validate(self) -> list[Recommendation]:  # Replace with actual return type
        # Implement validation logic here
        return [
            Recommendation(
                message="This is a dummy rule. Replace with actual validation logic.",
                code=self.code,
                fix="Toolkit team to add an actual rule",
            )
        ]



"""GraphQL Data Models API for managing CDF GraphQL/DML data models.

This API provides a wrapper around the legacy DML API for managing GraphQL data models.
"""

from collections.abc import Iterable, Sequence
from typing import Any

from pydantic import BaseModel, ConfigDict, Field, JsonValue

from cognite_toolkit._cdf_tk.client.cdf_client import CDFResourceAPI, Endpoint, PagedResponse
from cognite_toolkit._cdf_tk.client.http_client import (
    HTTPClient,
    ItemsSuccessResponse,
    RequestMessage,
    SuccessResponse,
    ToolkitAPIError,
)
from cognite_toolkit._cdf_tk.client.request_classes.filters import DataModelFilter
from cognite_toolkit._cdf_tk.client.request_classes.graphql import UPSERT_BODY
from cognite_toolkit._cdf_tk.client.resource_classes.data_modeling import DataModelReference
from cognite_toolkit._cdf_tk.client.resource_classes.graphql_data_model import (
    GraphQLDataModelRequest,
    GraphQLDataModelResponse,
)
from cognite_toolkit._cdf_tk.utils import humanize_collection


class UpsertResponseData(BaseModel):
    errors: dict[str, Any] | None = None
    result: GraphQLDataModelResponse


class GraphQLUpsertResponse(BaseModel):
    upsert_graph_ql_dml_version: UpsertResponseData = Field(alias="upsertGraphQlDmlVersion")


class GraphQLErrors(BaseModel):
    model_config = ConfigDict(extra="allow")
    message: str | None = None
    locations: list[dict[str, int]] | None = None
    extensions: dict[str, JsonValue] | None = None


class GraphQLResponse(BaseModel):
    data: GraphQLUpsertResponse
    errors: list[GraphQLErrors] | None = None


class GraphQLDataModelsAPI(CDFResourceAPI[GraphQLDataModelResponse]):
    """API for managing CDF GraphQL/DML data models.

    This API uses GraphQL mutations to manage data models with DML (Data Modeling Language).
    """

    def __init__(self, http_client: HTTPClient) -> None:
        super().__init__(
            http_client=http_client,
            method_endpoint_map={
                "create": Endpoint(method="POST", path="/dml/graphql", item_limit=1),
                "retrieve": Endpoint(method="POST", path="/models/datamodels/byids", item_limit=100),
                "delete": Endpoint(method="POST", path="/models/datamodels/delete", item_limit=100),
                "list": Endpoint(method="GET", path="/models/datamodels", item_limit=1000),
            },
        )

    def _validate_page_response(
        self, response: SuccessResponse | ItemsSuccessResponse
    ) -> PagedResponse[GraphQLDataModelResponse]:
        return PagedResponse[GraphQLDataModelResponse].model_validate_json(response.body)

    def _post_graphql(self, payload: dict[str, Any]) -> GraphQLUpsertResponse:
        """Execute a GraphQL query against the DML endpoint."""
        endpoint = self._method_endpoint_map["create"]
        request = RequestMessage(
            endpoint_url=self._make_url(endpoint.path),
            method=endpoint.method,
            body_content=payload,
        )
        result = self._http_client.request_single_retries(request)
        response = result.get_success_or_raise()
        parsed = GraphQLResponse.model_validate_json(response.body)
        if errors := parsed.errors:
            raise ToolkitAPIError(
                f"Failed GraphQL errors: {humanize_collection([error.message for error in errors if error.message])}"
            )
        return parsed.data

    def create(self, items: Sequence[GraphQLDataModelRequest]) -> list[GraphQLDataModelResponse]:
        """Apply (create or update) GraphQL data models in CDF.

        Args:
            items: List of GraphQLDataModelRequest objects to apply.

        Returns:
            List of applied GraphQLDataModelResponse objects.
        """
        results: list[GraphQLDataModelResponse] = []
        for item in items:
            payload = {
                "query": UPSERT_BODY,
                "variables": {"dmCreate": item.model_dump(mode="json", by_alias=True, exclude_unset=False)},
            }
            response = self._post_graphql(payload)
            results.append(response.upsert_graph_ql_dml_version.result)
        return results

    def retrieve(
        self, items: Sequence[DataModelReference], inline_views: bool = False
    ) -> list[GraphQLDataModelResponse]:
        """Retrieve GraphQL data models from CDF.

        Args:
            items: List of DataModelReference objects to retrieve.
            inline_views: Whether to include full view definitions in the response.

        Returns:
            List of retrieved GraphQLDataModelResponse objects.
        """
        return self._request_item_response(items, method="retrieve", extra_body={"inlineViews": inline_views})

    def delete(self, items: Sequence[DataModelReference]) -> None:
        """Delete GraphQL data models from CDF.

        Args:
            items: List of DataModelReference objects to delete.
        """
        self._request_no_response(items, "delete")

    def paginate(
        self,
        filter: DataModelFilter | None = None,
        limit: int = 100,
        cursor: str | None = None,
    ) -> PagedResponse[GraphQLDataModelResponse]:
        """Get a page of GraphQL data models from CDF.

        Args:
            filter: DataModelFilter to filter data models.
            limit: Maximum number of data models to return.
            cursor: Cursor for pagination.

        Returns:
            PagedResponse of GraphQLDataModelResponse objects.
        """
        return self._paginate(
            cursor=cursor,
            limit=limit,
            params=filter.dump() if filter else None,
        )

    def iterate(
        self,
        filter: DataModelFilter | None = None,
        limit: int | None = None,
    ) -> Iterable[list[GraphQLDataModelResponse]]:
        """Iterate over all GraphQL data models in CDF.

        Args:
            filter: DataModelFilter to filter data models.
            limit: Maximum total number of data models to return.

        Returns:
            Iterable of lists of GraphQLDataModelResponse objects.
        """
        return self._iterate(
            limit=limit,
            params=filter.dump() if filter else None,
        )

    def list(
        self,
        filter: DataModelFilter | None = None,
        limit: int | None = None,
    ) -> list[GraphQLDataModelResponse]:
        """List all GraphQL data models in CDF.

        Args:
            filter: DataModelFilter to filter data models.
            limit: Maximum total number of data models to return.

        Returns:
            List of GraphQLDataModelResponse objects.
        """
        return self._list(limit=limit, params=filter.dump() if filter else None)



import contextlib
from typing import Annotated

import typer

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.commands import InitCommand
from cognite_toolkit._cdf_tk.utils.auth import EnvironmentVariables


class LandingApp(typer.Typer):
    def __init__(self, *args, **kwargs) -> None:  # type: ignore
        super().__init__(*args, **kwargs)

    def main_init(
        self,
        dry_run: Annotated[
            bool,
            typer.Option(
                "--dry-run",
                "-r",
                help="Whether to do a dry-run, do dry-run if present.",
            ),
        ] = False,
    ) -> None:
        """Getting started checklist"""
        client: ToolkitClient | None = None
        with contextlib.redirect_stdout(None), contextlib.suppress(Exception):
            # Remove the Error message from failing to load the config
            # This is verified in check_auth
            client = EnvironmentVariables.create_from_environment().get_client()

        cmd = InitCommand(client=client)
        # Tracking  command with the usual lambda run construct
        # is intentionally left out because we don't want to expose the user to the warning
        # before they've had the chance to opt in (which is something they'll do later using this command).
        cmd.execute(dry_run=dry_run)



"""Scope definitions for Group capabilities.

Based on the API specification at:
https://api-docs.cognite.com/20230101/tag/Groups/operation/createGroups
"""

from typing import Annotated, Any, Literal, TypeAlias

from pydantic import BeforeValidator, Field, TypeAdapter, field_validator

from cognite_toolkit._cdf_tk.client._resource_base import BaseModelObject
from cognite_toolkit._cdf_tk.client.resource_classes.group._constants import SCOPE_NAME
from cognite_toolkit._cdf_tk.utils._auxiliary import get_concrete_subclasses


class ScopeDefinition(BaseModelObject):
    """Base class for all scope definitions."""

    scope_name: str


class AllScope(ScopeDefinition):
    """Scope that applies to all resources."""

    scope_name: Literal["all"] = Field("all", exclude=True)


class CurrentUserScope(ScopeDefinition):
    """Scope that applies to the current user only."""

    scope_name: Literal["currentuserscope"] = Field("currentuserscope", exclude=True)


class DataSetScope(ScopeDefinition):
    """Scope limited to specific data sets by ID."""

    scope_name: Literal["datasetScope"] = Field("datasetScope", exclude=True)
    ids: list[int]


class IDScope(ScopeDefinition):
    """Scope limited to specific resource IDs."""

    scope_name: Literal["idScope"] = Field("idScope", exclude=True)
    ids: list[int]


class IDScopeLowerCase(ScopeDefinition):
    """Scope limited to specific resource IDs (lowercase variant)."""

    scope_name: Literal["idscope"] = Field("idscope", exclude=True)
    ids: list[int]


class SpaceIDScope(ScopeDefinition):
    """Scope limited to specific spaces by ID."""

    scope_name: Literal["spaceIdScope"] = Field("spaceIdScope", exclude=True)
    space_ids: list[str]


class AssetRootIDScope(ScopeDefinition):
    """Scope limited to assets under specific root assets."""

    scope_name: Literal["assetRootIdScope"] = Field("assetRootIdScope", exclude=True)
    root_ids: list[int]


class TableScope(ScopeDefinition):
    """Scope limited to specific RAW tables."""

    scope_name: Literal["tableScope"] = Field("tableScope", exclude=True)
    dbs_to_tables: dict[str, list[str]]

    @field_validator("dbs_to_tables", mode="before")
    @classmethod
    def standardize_format(cls, value: Any) -> dict[str, list[str]]:
        """The API returns the dbsToTables field in a different format
        than the documentation specifies. This validator standardizes the format to match the documentation."""
        if not isinstance(value, dict):
            # Let pydantic handle the type error
            return value
        standardized: dict[str, list[str]] = {}
        for db, tables in value.items():
            if isinstance(tables, list):
                standardized[db] = tables
            elif isinstance(tables, dict) and "tables" in tables and isinstance(tables["tables"], list):
                standardized[db] = tables["tables"]
            elif tables == {}:
                standardized[db] = []
            else:
                raise ValueError(
                    f"Invalid format for dbsToTables: expected dict[str, list[str]] or dict[str, dict[tables: list[str]]], got {type(tables).__name__} for db '{db}'"
                )
        return standardized


class ExtractionPipelineScope(ScopeDefinition):
    """Scope limited to specific extraction pipelines."""

    scope_name: Literal["extractionPipelineScope"] = Field("extractionPipelineScope", exclude=True)
    ids: list[int]


class InstancesScope(ScopeDefinition):
    """Scope limited to specific instances."""

    scope_name: Literal["instancesScope"] = Field("instancesScope", exclude=True)
    instances: list[str]


class PartitionScope(ScopeDefinition):
    """Scope limited to specific partitions."""

    scope_name: Literal["partition"] = Field("partition", exclude=True)
    partition_ids: list[int]


class ExperimentScope(ScopeDefinition):
    """Scope limited to specific experiments."""

    scope_name: Literal["experimentscope"] = Field("experimentscope", exclude=True)
    experiments: list[str]


class AppConfigScope(ScopeDefinition):
    """Scope limited to specific app configurations."""

    scope_name: Literal["appScope"] = Field("appScope", exclude=True)
    apps: list[str]


class PostgresGatewayUsersScope(ScopeDefinition):
    """Scope limited to specific PostgreSQL gateway users."""

    scope_name: Literal["usersScope"] = Field("usersScope", exclude=True)
    usernames: list[str]


class UnknownScope(ScopeDefinition):
    """Fallback class for unknown scope definitions."""

    scope_name: str


def _get_scope_name(cls: type[ScopeDefinition]) -> str | None:
    """Get the scope_name default value from a Pydantic model class."""
    field = cls.model_fields.get("scope_name")
    if field is not None and field.default is not None:
        return field.default
    return None


_KNOWN_SCOPES = {
    name: scope
    for scope in get_concrete_subclasses(ScopeDefinition)
    if (name := _get_scope_name(scope)) is not None and scope is not UnknownScope
}


def _handle_unknown_scope(value: Any) -> Any:
    if isinstance(value, dict) and isinstance(scope_name := value.get(SCOPE_NAME), str):
        scope_class = _KNOWN_SCOPES.get(scope_name)
        if scope_class:
            return TypeAdapter(scope_class).validate_python(value)
    return UnknownScope.model_validate(value)


Scope: TypeAlias = Annotated[
    (
        AllScope
        | CurrentUserScope
        | DataSetScope
        | IDScope
        | IDScopeLowerCase
        | SpaceIDScope
        | AssetRootIDScope
        | TableScope
        | ExtractionPipelineScope
        | InstancesScope
        | PartitionScope
        | ExperimentScope
        | AppConfigScope
        | PostgresGatewayUsersScope
        | UnknownScope
    ),
    BeforeValidator(_handle_unknown_scope),
]
